#include <unistd.h>
#include "init.h"
#include <signal.h>
#include "icmp_util.h"
#include "client_icmp.h"
#include "cl_messages.h"
#include "util.h"


#define SOCK_BUFF_LEN 64000
#define TRUE 1
#define FALSE 0
#define TTL 255
#define Mbps  1048576           //Megabit -> bit

//#define DEBUG

struct icmphdr *ih;
struct timeval *send_time, *ret_time, *result_time;
//parametry *param;
ICMP_STR *send_str;	
int finished;
double time_trip;  /*doba cesty 1 ICMP paketu*/
FILE * outf2;

void icmp_cleanup(int sigNo) {
  if ( sigNo != 0) {
    finished=2;
  } else {
    finished=TRUE;
  }
  ICMP_alarm(0);
  signal(SIGALRM, SIG_IGN);
  do_send=FALSE;
}

void get_icmp_time(int sigNo){
  /*cli_serv = (((serv_in_time_s+ (double)serv_in_time_us/1000000) - (cli_out_time_s + (double)cli_out_time_us/1000000)))/2;
  */
  result_time->tv_sec=ret_time->tv_sec - send_time->tv_sec;
  result_time->tv_usec=ret_time->tv_usec - send_time->tv_usec;
  time_trip=(result_time->tv_sec + (double)result_time->tv_usec/1000000);

#ifdef DEBUG
  printf("send %d %d\n",send_time->tv_sec,send_time->tv_usec);
  printf("ret  %d %d\n",ret_time->tv_sec,ret_time->tv_usec);
  printf("around  %f \n",time_trip);
#endif  	
};
/*Handler pro odesilani paketu*/
/* ------------send_icmp------------*/
void send_icmp(int sigNo){
  int send;
#ifdef DEBUG
  printf("sock=%d\n",send_str->sock);
  printf("srv_addr=%s\n",send_str->srv_addr);
  printf("data=%s\n",send_str->data);
  printf("length=%d\n",send_str->param->length+ICMP_HDR_MIN);
#endif
  send=ICMP_echo_send(send_str->sock,
                      send_str->srv_addr,
                      send_str->data,
                      send_str->param->length+ICMP_HDR_MIN,
                      TTL,
                      NULL);
  if (send < SND_ERR_OFS) {
    LOG_ERR("ICMP","Chyba v odesilani ICMP paketu");
    fprintf(outf2,"Chyba v odesilani ICMP paketu: %s<BR>\n", 
		      strerror(-send + SND_ERR_OFS));
    fflush(outf2);
    icmp_cleanup(-1);
    return;
  }
  if(++ih->un.echo.sequence==send_str->param->count) {
    icmp_cleanup(0);
  }
#ifdef DEBUG
  printf("send %d, send=%d\n",ih->un.echo.sequence,send);
#endif
}


/*funkce nastavuje handler na odesilani paketu*/
/* ------------start_icmp------------*/
void start_icmp(){
  signal(SIGALRM, send_icmp);
  signal(SIGUSR1, get_icmp_time);
  signal(SIGUSR2, icmp_cleanup);
}


/*funkce nastavuje ICMP hlavicku*/
/* ------------set_icmp_stream------------*/
struct icmphdr *set_icmp_stream(char *data){
  struct icmphdr *ih1;
  srand(time(NULL));
  ih1=(struct icmphdr *)data;
  send_str->stream=ih1->un.echo.id = rand();	/*id icmp streamu*/
  ih1->un.echo.sequence = 0;                    /*id icmp paketu ve streamu*/
  return ih1;
}


/* ------------protocolICMP------------*/
int protocolICMP(parametry *param, FILE* outf) {
  struct sockaddr_in srv_adr, cli_adr;                  /* Adresa pro bind   */
  int i,recv,ttl_get,sequence;
  char data_in [SOCK_BUFF_LEN];                             /* Prichozi buffer   */
//  struct timeval ret_time;                              /* Cas prichodu      */
  struct hostent * hostinf = NULL;                      /* Info od resolveru */
  double casSUM=0;

  LOG_ERR_S("ICMP-info","Start na server",param->IP);
  /*Nastaveni run-time struktury*/
  send_str = (ICMP_STR *) malloc(sizeof(ICMP_STR));     /* Alokovani struktury*/
  send_str->param=param;
  send_str->data = malloc(param->length + MIN_ECHO);
  ret_time=(struct timeval*) malloc(sizeof(struct timeval));
  send_time=(struct timeval*) malloc(sizeof(struct timeval));
  result_time=(struct timeval*) malloc(sizeof(struct timeval));
  outf2=outf;

#ifdef DEBUG
    printf("Struktura prikazove radky:\n");
    printf("\tprotocol=%s\n",send_str->param->protocol);
    printf("\tcount=%d\n",send_str->param->count);
    printf("\tlength=%d\n",send_str->param->length);
    printf("\tbuffsocket=%d\n",send_str->param->buffSocket);
    printf("\tIP=%s\n",send_str->param->IP);
    printf("\tfrekvence=%d\n",send_str->param->freq);
    printf("\tvstup_data=%s\n",send_str->param->vstup_data);
    printf("\tadr=%s\n",send_str->srv_addr);

#endif

  srv_adr.sin_family = AF_INET; 			/* Trida adres       */

    /** Vzdaleny: hostname, IP adresa */
  if (inet_aton(send_str->param->IP, &srv_adr.sin_addr) != 1) {
    if ((hostinf = gethostbyname(send_str->param->IP)) == 0) {
      LOG_ERR("PING", "Nelze zjistit IP adresu hostu");
      fprintf(outf,"Nelze zjistit IP adresu hostu");
//      fprintf(outf, M_IP_NOTFOUND, p->host);
//      fflush(outf);
//      ping_free(NULL,ps);
      return 0;
    }
    free(send_str->param->IP);
    strncpy(send_str->param->IP,hostinf->h_name,4);
    memcpy(&srv_adr.sin_addr, hostinf->h_addr_list[0], 4); /* Kopie adresy   */
  }
  send_str->srv_addr=&srv_adr;
  send_str->sock=ICMP_open(&cli_adr); 				/*Otevremni socketu  */
  if (send_str->sock < 0) {  					/* Chyba pri registraci */
    fprintf(outf, "Nelze zacit (nedostatecna privilegia?)\n");
    return 0;
  }
    /* Zakazat fragmentaci */
  i = IP_PMTUDISC_DO;
  if (setsockopt(send_str->sock, SOL_IP, IP_MTU_DISCOVER, &i, sizeof(i))) {
     fprintf(stderr, "Nelze zakazat fragmentaci.\n");
  }

  /*Priprava dat*/
  ICMP_echo_prep(send_str->data, send_str->param->length+ICMP_HDR_MIN, send_str->param->vstup_data);
  /*Nastaveni icmp hlavicky*/
  ih=set_icmp_stream(send_str->data);
#ifdef DEBUG
  printf("icmp stream=%d icmp id=%d\n",ih->un.echo.id,ih->un.echo.sequence);
#endif
  /*Nastaveni frekvence posilani icmp paketu*/
  ICMP_alarm_us(send_str->param->freq);
  start_icmp();
  finished = FALSE;
  do_send=TRUE;
  sequence=0;
  send_icmp(0);
  
  /*   Nacitani ICMP echo replay   */
  while (finished == FALSE) {
     recv=ICMP_echo_recv(send_str->sock,
  		        send_str->srv_addr,
                        (char *) &data_in,
                        SOCK_BUFF_LEN,
                        (char *)&ttl_get,
                        send_str->stream,
                        ret_time
  		        );

    if (recv <= FAILED_CRC) { continue; }
    if (recv < 0 && recv > FAILED_CRC ) {
      fprintf(outf,"Chybova ICMP zprava: %s<BR>\n",
		                ICMP_message((char*)&data_in));
    } else if (ICMP_echo_seqNo(data_in) >= sequence) {
       * send_time = * (struct timeval *) (data_in + ICMP_HDR_MIN);
       kill(getpid(),SIGUSR1);	
       casSUM+=time_trip;
       fprintf(outf,ICMP_VYPIS_O_PAKETU,ICMP_echo_seqNo(data_in),
	  (float)time_trip*1000,
          (2*send_str->param->length/(float)time_trip)*8/Mbps,
          ICMP_message((char*)&data_in));
       fflush(outf);
       sequence=ICMP_echo_seqNo(data_in)+1;
    }
//    printf("sequence=%d return sequence=%d\n",sequence,ih->un.echo.sequence);
  }//end while

  signal(SIGUSR2, SIG_IGN);
  signal(SIGUSR1, SIG_IGN);
  signal(SIGALRM, SIG_IGN);
  if (finished == TRUE && sequence) {
    fprintf(outf,ICMP_STATISTICS, send_str->param->count,
	    send_str->param->length,
	    send_str->param->count*send_str->param->length,
	    (float)time_trip,
	    ((send_str->param->count*send_str->param->length*2)/(float)casSUM)*8/Mbps);
    fflush(outf);
  }
  free(send_str->data);
  free(send_str);
  free(ret_time);
  free(send_time);
  free(result_time);
  LOG_ERR("ICMP-info","Uspesne ukonceno");
  return 1;
}
