#include "init.h"
#include "funkce.h"
#include "icmp_util.h"
#include "util.h"
#include "mutils.h"
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/ioctl.h>

#define UDP_IP_HDR 28 /* Min. delka IP+UDP hlavicky */
#define MINFREQ 500 /* 500 us */
#define TIMEOUT 2	/* nastaveni timeoutu pro naslouchani na socketu */
#define MAX_TOUTS 3	/* po kolika za sebou nasledujicich timeoutech 
			                                    vyhlasi konec */
#define MIN_USECS 100   /*Pokud by kvuli nepresnemu mereni byl cas <= 0 */
#define Mbps  1048576   /* Megabit -> bit */

#define  MBPS_SPEED(ts, tt)  ((ts) * 8) / (Mbps * ((tt)/1000000.0F))

int q, i;
struct timeval tvpom, timeout;

struct timeval * tv = &tvpom;

struct timeval tzero = {0, 0};

int running = TRUE; 	/* Prijem bezi */
int chpid; 		/* Synovsky proces */

void protocolUDP2_stop(int sigNo) {
  running = FALSE;
  if (chpid) { kill(chpid, SIGUSR2); }	  
  signal(SIGUSR2, SIG_IGN);
}	

int protocolUDP2_sender(parametry *param, FILE *outf, int sockfdudp, 
		                                    struct sockaddr_in * srv ) {
  int i, sel, first = 1;	
  int lost=0;
  int tout_cnt=0;
  int lenUDP;	   /* pocet odeslanych bytu */
  struct timeval t_beg, t_end;
  struct timeval diff_beg, diff_end;
  long long total_snd_time;

  char *buffer;
  fd_set wfds; /* struktura pro popis struktury FileDes. */
  
  buffer = (char *) malloc((param->length+1) * sizeof(char));
  fillBuffer(buffer + UDP_ALL, param->length-UDP_ALL, param->vstup_data);
  
  gettimeofday(&t_beg, NULL);
  diff_beg = t_beg;
  
  for (i=0; i < param->count; i++) {
     int j = param->count - i - 1, k = i+1;   
     if (! running) { break; }
     
     memcpy(buffer+UDP_COUNT, &j, sizeof(j));
     memcpy(buffer+UDP_CURR, &k, sizeof(k));
     FD_ZERO(&wfds);
     FD_SET(sockfdudp, &wfds);
     timeout.tv_sec=TIMEOUT;
     timeout.tv_usec=0;
     
     gettimeofday(&diff_beg, NULL); /* Zacatek intervalu */
     
     do {
       sel=select(sockfdudp+1, NULL, &wfds, NULL, &timeout);
     } while (sel == -1 && errno == EINTR);
     
     gettimeofday(&diff_end, NULL); /* Konec intervalu */
     tv_sub(&diff_beg, &diff_end);  /* Zaporna hodnota */
     tv_sub(&t_beg, &diff_beg);     /* Posuneme zacatek - 
				                 uprava delky intervalu */
     switch(sel){
     
	  case -1:   LOG_ERR_P("UDP", "Chyba v selektu odesilatele"); 
		     fprintf(outf,"Chyba v selektu odesilatele<BR>\n");
		     lost += param->count - i;
		     running = FALSE; 
		     break;
	  case  0:   LOG_ERR("UDP", "Neodeslano");
		     if (++tout_cnt == MAX_TOUTS) {
                       fprintf(outf, "Odesilani ukonceno: cas vyprsel<BR>\n");
		       running=FALSE;
		       lost += param->count - i;
		       continue;
	             } else {
		       lost++;	     
	             }		     
		     break;
		     
	  case  1:   /* Cas odeslani do paketu */ 
		     tout_cnt = 0;
		     gettimeofday((struct timeval*)(buffer+CLIENT_TIME), NULL);
	             lenUDP = sendto(sockfdudp, buffer, 
				     param->length, 0, (struct sockaddr *)srv,
				                                  sizeof(*srv));
	             //printf("odeslan udp %d z %d\n", lenUDP, param->count);
	             break;
     }
     gettimeofday(&diff_beg, NULL); /* Zacatek intervalu */
	
     if (lenUDP<0) {
       if (first) {
	 first=0;
         LOG_ERR("UDP", "chyba pri odesilani UDP paketu");
         fprintf(outf, "<H2>Chyby pri odesilani UDP paketu</H2>\n");
       }
       lost++;
       fprintf(outf, "Nepodarilo se odeslat paket c. %i<BR>\n", k);
     }

     usleep(param->freq);           /* At dalsi proces posila data */
     gettimeofday(&diff_end, NULL); /* Konec intervalu */
     tv_sub(&diff_beg, &diff_end);  /* Zaporna hodnota */
     tv_sub(&t_beg, &diff_beg);     /* Posuneme zacatek - 
				                 uprava delky intervalu */
  }
  
  gettimeofday(&t_end, NULL);
  tv_sub(&t_end, &t_beg);
  total_snd_time = tv2usecs(&t_end);

  fprintf(outf, "<H2>Statistika odeslanych UDP paketu</H2>\n");
  
  if (lost==param->count) {
     fprintf(outf, "<BR>Nic nebylo odeslano<BR>\n");
     free(buffer);
     return 0;
  }
    
  fprintf(outf,	"<PRE>\n"
		"\tPocet paketu:\t %d ks\n"
		"\tDelka paketu:\t %d B\n"
		"\tOdeslano:\t %d B\n"
		"\tRychlost:\t %.3f Mbps\n"
                "\tDoba odesilani:\t %s\n"
		"</PRE>\n", 
		param->count-lost, param->length, 
		(param->count - lost) * param->length,
		MBPS_SPEED(param->count*param->length, total_snd_time),
 		usecs2str(total_snd_time)
  );		

  free(buffer);
  return param->count - lost;
}
	
/**
 * @param count      Pocet paketu
 * @param length     Delka paketu
 * @param data       Data, kterymi bude paket vyplnen
 * @param IP         IP adresa, kam data posilat
 * @param buffSocket Velikost bufferu socketu
 * @param vystup     Vystup
 */
int protocolUDP2(parametry *param, FILE *outf){
    char *buffer;
    struct sockaddr_in serv_addr;
    int lenUDP;	   /* pocet prectenych bytu ze vstupu */
    int sockfdudp;
    int srvlen;
    int sel, lost=0;
    struct timeval timeout, t1, t2, t3, t4, t5, travel;
    fd_set rfds; /* struktura pro popis struktury FileDes. */
    char * s1, *s2, *s3, *s4;
    long long ttime, tot_time=0, min_time=-1, max_time=-1, avg_time=0;
    int pipa[2], c;
    int tout_cnt=0;
    FILE * inf;
    
    /*if (param->freq<MINFREQ)*/ { param->freq=MINFREQ; }
    
    chpid = 0;
    running = TRUE; 	
    signal(SIGUSR2, protocolUDP2_stop); 
    
    fflush(outf);
    LOG_ERR_S("UDP-info",  "Start na server",param->IP);

    if ((sockfdudp = socket(AF_INET, SOCK_DGRAM, 0)) < 0) { 
       /* otevreni socketu pro UDP */
       LOG_ERR("UDP", "Nemohu otervrit socket");
       fprintf(outf, "Nemohu otervrit socket<BR>\n");
       return -1;
    }

    setsockopt(sockfdudp, SOL_SOCKET, SO_SNDBUF, 
		    		&param->buffSocket, sizeof(param->buffSocket));
    setsockopt(sockfdudp, SOL_SOCKET, SO_RCVBUF, 
		    		&param->buffSocket, sizeof(param->buffSocket));

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(param->IP);
    serv_addr.sin_port = htons(param->port);
    srvlen =  sizeof(serv_addr);

    buffer = (char *) malloc((param->length+1) * sizeof(char));

    fcntl(sockfdudp, F_SETFL, O_NONBLOCK);
	
    if (pipe(pipa) < 0) {
        LOG_ERR_P("UDP", "Nelze otevrit pipu vysledku");
    }	    
    switch (chpid = fork()) {
       case  -1: 
	         close(pipa[1]); close(pipa[0]);
	         LOG_ERR_P("UDP", "Nelze spustit odesilaci cast");
		 close(sockfdudp);
                 free(buffer);
                 signal(SIGUSR2, SIG_IGN);
                 return -1;
       case   0:
		 outf = fdopen(pipa[1], "w");
	         close(pipa[0]); 
		 if (protocolUDP2_sender(param, outf, 
					        sockfdudp, &serv_addr) < 0) {
                    kill(getppid(), SIGUSR2);
		 }
		 fclose(outf);
		 exit(0);
       default: 
	         close(pipa[1]); 
		 inf = fdopen(pipa[0], "r");
    }	    
    
    fprintf(outf, "<H2>Prijate UDP pakety</H2>\n");
    for (i=0; i < param->count; i++) {
        if (! running) { break; }

        timeout.tv_sec=TIMEOUT;
        timeout.tv_usec=0;
    
	FD_ZERO(&rfds);
	FD_SET(sockfdudp, &rfds);
	
	do {
	  sel=select(sockfdudp+1, &rfds, NULL, NULL, &timeout);
        } while (sel == -1 && errno == EINTR);
 		
	switch(sel){
	  case -1:   LOG_ERR_P("UDP", "Chyba v selektu prijemce"); 
		     fprintf(outf,"Chyba v selektu prijemce<BR>\n");
		     lost += param->count - i;
		     running = FALSE; 
		     break;
		     
  	     /*kdyz do TIMEOUT sekund se nic nestane, tzn: neprichazi odpoved*/
	  case  0:   //error("Nic neprislo\n");
		     if (++tout_cnt == MAX_TOUTS) {
                       fprintf(outf, "Prijem ukoncen: cas vyprsel<BR>\n");
		       lost += param->count - i;
		       running=FALSE;
		       continue;
	             } else {		     
  	               lost++;
		     }
		     continue;
		     
	  case  1:   /* Ke klientovi */
		     tout_cnt = 0;
		     gettimeofday(&t5, NULL);
	             lenUDP = recvfrom(sockfdudp, buffer, param->length, 0,
				                (struct sockaddr *)&serv_addr, 
						                       &srvlen);
		     if (ioctl(sockfdudp, SIOCGSTAMP, &t4) < 0) { t4 = t5; }	     
	             if (lenUDP==-1) {
	                lost++;
	                continue;
	             }

		     /* Od klienta */
	             t1 = * (struct timeval*) (buffer+CLIENT_TIME); 
	
		     /* K serveru */
	             t2 = * (struct timeval*) (buffer+SERVER_TIME);
		     
		     /* Od serveru */
	             t3 = * (struct timeval*) (buffer+SERVER_TIME_OUT); 
		     
		     tv_sub(&t3, &t2);
		     tv_sub(&t4, &t1);
		     
		     timeval_subtract(&travel, &t4, &t3); 
		     ttime=tv2usecs(&travel);
		     if (ttime <= 0) { ttime = MIN_USECS; }

		     fprintf(outf, "%3i: celkem: %s, ", 
				                * (int*)(buffer+UDP_CURR),
				                  usecs2str(tv2usecs(&t4)));  

		     fprintf(outf, "na serveru: %s, ",
				      		usecs2str(tv2usecs(&t3)));  

                     fprintf(outf, "rtt: %s, rychlost: %8.3f Mbps<BR>\n", 
				   usecs2str(ttime),  
				   MBPS_SPEED(param->length, ttime / 2/*???*/)
                     );
		     fflush(outf);
		     
		     ttime /= 2;  /* Vydelime 2 */ /* ??? */

		     if (min_time < 0 || ttime < min_time) {
		       min_time = ttime;
		     }	     
		     
		     if (max_time < 0 || ttime > max_time) {
		       max_time = ttime;
		     }	     
		     
		     tot_time += ttime;	     
		     
	             break;
	} //end switch
    } //end for
    avg_time = (param->count == lost)? 0: (tot_time / (param->count - lost));

    waitpid(chpid, NULL, 0);
    chpid=0;
    signal(SIGUSR2, SIG_IGN); 

    LOG_ERR("UDP-info", "Uspesne ukonceno");
    
    fflush(outf);
    while ((c = fgetc(inf)) > -1) { fputc(c, outf); }
    fclose(inf);
    fflush(outf);
    
    fprintf(outf, "<H2>Statistika prijatych UDP paketu</H2>\n");
    
    /*kvuli statistice, kdyz se strati vse*/
    if (lost==param->count) {
       fprintf(outf, "<BR>Nic nebylo prijato<BR>\n");
       free(buffer);
       return 0;
    }
    
    fprintf(outf, "<PRE>\n"
		  "\tPocet paketu:\t %d ks\n"
		  "\tDelka paketu:\t %d B\n"
		  "\tPreneseno:\t %d B\n"
		  "\tZtraceno:\t %d ks %2.2f%%\n"
         	  "\tDoba prenosu:\t %s\n"
		  "\tRychlost:\t %.3f Mbps\n\n"
                  "\tDoba prenosu (rtt/2): - minimalni:\t %s\n"
		  "\t\t\t      - prumerna:\t %s\n"
		  "\t\t\t      - maximalni:\t %s\n\n"
		  "\tRychlost: - minimalni:\t %0.6f Mbps\n" 
		  "\t\t  - prumerna:\t %0.6f Mbps\n"
                  "\t\t  - maximalni:\t %0.6f Mbps\n"
		  "</PRE>\n",
	    param->count - lost,
	    param->length,
	    (param->count-lost)*param->length,
	    lost, (float)lost*100/(float)param->count,
	    s1=strdup(usecs2str(tot_time)),
	    MBPS_SPEED((param->length/*+UDP_IP_HDR*/)*(param->count-lost),
	                                                       tot_time),
	    s2=strdup(usecs2str(min_time)),
	    s3=strdup(usecs2str(avg_time)),
	    s4=strdup(usecs2str(max_time)),
	    MBPS_SPEED(param->length/*+UDP_IP_HDR*/, max_time),
	    MBPS_SPEED(param->length/*+UDP_IP_HDR*/, avg_time),
	    MBPS_SPEED(param->length/*+UDP_IP_HDR*/, min_time)
    );	    
    free (s1); free (s2); free (s3); free (s4);
    free(buffer);
    return param->count - lost;
}
