#include "init.h"
#include "funkce.h"
#include "mutils.h"
#include "util.h"
#include "cl_messages.h"
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#define DELAY_TCP  0000000L
#define DELAY_UDP  0000000L
#define ICMPDATALEN 56
#define TIMEOUT 5		//nastaveni timeoutu pro naslouchani na socketu
#define Mbps  1048576           //Megabit -> bit

int runningTCP = TRUE;     /* Prijem bezi */

void protocolTCP_stop(int sigNo) {
 runningTCP = FALSE;
}


int q,i;
struct timeval *tv, tvpom,timeout;

//........................................................................
/**
 * @param count      Pocet paketu
 * @param length     Delka paketu
 * @param data       Data, kterymi bude paket vyplnen
 * @param IP         IP adresa, kam data posilat
 * @param buffSocket Velikost bufferu socketu
 * @param vystup     Vystup
 */
int protocolTCP(parametry *param, FILE * outf) {
    char * buffer;
    int sockfd,lenTCP,sel,suma,prijato;
    struct sockaddr_in serv_addr;
    long s1, us1, s2, us2;
    long totalLength;
    float cas,c_cas;
    fd_set rfds,wfds; //struktura pro popis struktury FileDes.

    tv = &tvpom;
    runningTCP = TRUE;
    signal(SIGUSR2, protocolTCP_stop);
	
    if (param->IP == NULL) { 
      param->IP=strdup("127.0.0.1");
      LOG_ERR("TCP","Spusteno testovani TCP s NULL adresou");
    }

        //vynuluje byty
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(param->IP);
    serv_addr.sin_port = htons(param->port);

    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
       LOG_ERR("TCP","Chyba pri 'socket'");
       fprintf(outf, TCP_SOCKET_CHYBA);
       signal(SIGUSR2, SIG_IGN);
       return -1;
    }

    setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &param->buffSocket, sizeof(param->buffSocket));

    LOG_ERR_S("TCP-info","Start na server",param->IP);

    if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
       LOG_ERR("TCP","Chyba pri 'connect'");
       fprintf(outf, TCP_CONNECT_CHYBA);
       signal(SIGUSR2, SIG_IGN);
       return -1;

    }
    buffer = (char *) malloc((param->length+1) * sizeof(char));
    
    totalLength = 0;
    fillBuffer(buffer, param->length, param->vstup_data);
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_SET(sockfd,&rfds);
    FD_SET(sockfd,&wfds);

    suma=0;   //pro ladici informace, pocet prijatych bytu
    c_cas=0;  //vynulovani celkoveho casu.
    for(i=0; i<param->count; i++) { //posilani dat
	if(runningTCP==FALSE){
          close(sockfd);
          free(buffer);
          signal(SIGUSR2, SIG_IGN);
          return -1;
        }
        timeout.tv_sec=TIMEOUT;
        timeout.tv_usec=0;

	sel=select(sockfd+1,NULL,&wfds,NULL,&timeout);
	switch(sel){
	    case -1: LOG_ERR("TCP","Chyba v selektu"); break;
	    case  0: break; //nic se nestalo
	    case  1: getTime(&s1, &us1);
	             lenTCP=write(sockfd, buffer, param->length);
                     if(lenTCP!=param->length){
                       LOG_ERR_S("TCP","Nepovedlo se zapsat vsechna data do socketu",strerror(errno));
                     }
	             suma+=lenTCP;
#ifdef DEBUG			
	             printf("odeslan paket o velikosti %d bytu\n",lenTCP);
#endif	
	             break;
	}//end switch
	prijato=lenTCP;
	while(prijato){
          FD_SET(sockfd,&rfds);
          timeout.tv_sec=TIMEOUT;
          timeout.tv_usec=0;
	  
	  sel=select(sockfd+1,&rfds,NULL,NULL,&timeout);
          switch(sel){
	      case -1: LOG_ERR("TCP","Chyba v selektu"); break;
	      case  0: break; //nic se nestalo
	      case  1: getTime(&s2, &us2);
	               lenTCP = read(sockfd, buffer, param->length);
#ifdef DEBUG		
	               printf("paket o velikosti %d bytu\n",lenTCP);
#endif	
	               prijato=prijato-lenTCP;
	               break;
  	  }//end switch
        }//end while
        cas = ((s2 + (double)us2/1000000) - (s1+ (double)us1/1000000));
        c_cas+=cas;
        fprintf(outf,TCP_VYPIS_O_PAKETU,i+1,cas*1000,c_cas,
           (param->length*2/cas)*8/Mbps);
        fflush(outf);
      } //end for
#ifdef DEBUG	
    printf("suma=%d\n",suma);
#endif
    close(sockfd);
    LOG_ERR("TCP-info","Uspesne ukonceno");
    fprintf(outf, TCP_STATISTICS,
	    param->count,
	    param->length,
	    param->count*param->length,
	    c_cas, 
	    ((param->count*param->length*2)/c_cas)*8/Mbps);
    fflush(outf);
	    
    free(buffer);
    signal(SIGUSR2, SIG_IGN);
    return 1;
}


/*
 * protocolUDP byl prepsan na jinou variantu protocolUDP2 do souboru 
 * client_udp.c, ktera se nyni uziva. Ponechano jen pro mozny navrat k puvodni 
 * verzi
 */
#ifdef NOTDEF
/**
 * @param count      Pocet paketu
 * @param length     Delka paketu
 * @param data       Data, kterymi bude paket vyplnen
 * @param IP         IP adresa, kam data posilat
 * @param buffSocket Velikost bufferu socketu
 * @param vystup     Vystup
 */
int protocolUDP(parametry *param, FILE *outf){
    char *buffer;
    struct sockaddr_in serv_addr, cli_addr;
    int serv_len;//,cli_len;
    int lenUDP; // pocet prectenych bytu ze vstupu
    int incoming;  //pocet paketu, ktere dosly
    int sockfdudp;//, newsockfdudp;
    float casPom;
    float casAvg, casSum, casMin, casMax, cas_serv, cas_cli;
    int sel,lost=0;

    struct timeval timeout,*t1,*t2,*t3,*t4;
    fd_set wfds,rfds; //struktura pro popis struktury FileDes.
    int cc;
    float SumLength;
    long cli_out_time_s,cli_out_time_us;
    long serv_in_time_s,serv_in_time_us;
    long serv_out_time_s,serv_out_time_us;
    long cli_in_time_s,cli_in_time_us;

    float cli_serv,serv_cli;


    tv = &tvpom;
//    if (param->count<1) { param->count=1; }
//    if (param->length<100) { param->length=100; }
//    if (param->length>64000) { param->length=64000; }

    printf("UDP - start\n");

       if ((sockfdudp = socket(AF_INET, SOCK_DGRAM, 0)) < 0) { //otevreni socketu pro UDP
       LOG_ERR("UDP","Nemohu otevrit socket");
       fprintf(outf, UDP_SOCKET_CHYBA);
       return -1;
    }
    setsockopt(sockfdudp, SOL_SOCKET, SO_SNDBUF, &param->buffSocket, sizeof(param->buffSocket));
    setsockopt(sockfdudp, SOL_SOCKET, SO_RCVBUF, &param->buffSocket, sizeof(param->buffSocket));

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(param->IP);
    serv_addr.sin_port = htons(param->port);

    bzero((char *) &cli_addr, sizeof(cli_addr));	/* zero out */
    cli_addr.sin_family      = AF_INET;
    cli_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    cli_addr.sin_port        = htons(0);

    serv_len = sizeof(serv_addr);

    buffer = (char *) malloc((param->length+1) * sizeof(char));

    casSum = cas_serv = cas_cli = 0;
    casMin = 99999; casMax = 0;    
//    getTime(&h1, &uh1);
    cc=0;
    incoming=0;
    fcntl(sockfdudp,F_SETFL,O_NONBLOCK);
    t1=(struct timeval*) malloc(sizeof(struct timeval));
    t2=(struct timeval*) malloc(sizeof(struct timeval));
    t3=(struct timeval*) malloc(sizeof(struct timeval));
    t4=(struct timeval*) malloc(sizeof(struct timeval));
	
    timeout.tv_sec=TIMEOUT;
    timeout.tv_usec=0;
    for (i=0; i<param->count;) {
	//****************************
        //if (i==0) bcopy(&param->count, buffer+UDP_COUNT,4);
        bcopy(&param->count, buffer+UDP_COUNT,4);
	
	FD_ZERO(&wfds);
        FD_SET(sockfdudp,&wfds);
        sel=select(sockfdudp+1,NULL,&wfds,NULL,&timeout);
	switch(sel){
	  case -1:   LOG_ERR("UDP","Chyba v selektu"); break;
	  case  0:   printf("Neodeslano\n");break;
	  case  1:   gettimeofday(tv,NULL);
	             bcopy((char*)tv, buffer+CLIENT_TIME, 8); //zapis cas odeslani paketu
	             lenUDP = sendto(sockfdudp, buffer, param->length, 0, &serv_addr, serv_len);
//	             printf("odeslan udp %d z %d\n",lenUDP,param->count);
	             //cc++; /* <-- pouze pro ladici informace */
	             i++;
	             break;
	}
	
	if (lenUDP<0) {
	    LOG_ERR("UDP","Chyba pri 'sendto'");
	    fprintf(outf, UDP_SENDTO_CHYBA);
	    return -1;
	    break;
      }
	FD_ZERO(&rfds);
	FD_SET(sockfdudp,&rfds);
	sel=select(sockfdudp+1,&rfds,NULL,NULL,&timeout);
	switch(sel){
	  case -1:   LOG_ERR("UDP","Chyba v selektu"); break;
	  	     /*kdyz do 5 sekund se nic nestane, tzn: neprichazi odpovedi*/
	  case  0:   //error("Nic neprislo\n");
	             lost++;
	             break;
	  case  1:   gettimeofday(t4,NULL);
	             lenUDP = recvfrom(sockfdudp, buffer, param->length, 0, &serv_addr, &serv_len);
	             if (lenUDP==-1) {
	                //incoming++;
	                //fprintf(outf,"Chyba pri cteni ze socketu<BR>\n");
	                lost++;
	                continue;
	                //return -1;
	             }
//	             i++;
	             incoming++;
#ifdef DEBUG1	
	             printf("prijaty %d paket vel %d\n",cc,lenUDP);
#endif	
	             cli_in_time_s  = t4->tv_sec;
	             cli_in_time_us = t4->tv_usec;

	             t1 = (struct timeval*) (buffer+CLIENT_TIME); //Odchod od clienta
	             cli_out_time_s  = t1->tv_sec;
	             cli_out_time_us = t1->tv_usec;
	
	             t2 = (struct timeval*) (buffer+SERVER_TIME);
	             serv_in_time_s  = t2->tv_sec;
	             serv_in_time_us = t2->tv_usec;

	             t3 = (struct timeval*) (buffer+SERVER_TIME_OUT); //Odchod od clienta
	             serv_out_time_s  = t3->tv_sec;
	             serv_out_time_us = t3->tv_usec;
	                 //klient -> server
	             cli_serv = (((serv_in_time_s+ (double)serv_in_time_us/1000000) - (cli_out_time_s + (double)cli_out_time_us/1000000)))/2;
	                 //server->klient
	             serv_cli = (((cli_in_time_s + (double)cli_in_time_us/1000000) - (serv_out_time_s+ (double)serv_out_time_us/1000000)))/2;
	             casPom = (cli_serv+serv_cli)/2;  //prumer tam i zpet jedne cesty paketu

#ifdef DEBUG	
	             printf("\ncli_in_time_s=%d  serv_in_time_s=%d\n",cli_in_time_s,serv_in_time_s);
	             printf("cli_in_time_us=%d  serv_in_time_us=%d\n",cli_in_time_us,serv_in_time_us);
	             printf("cli_out_time_s=%d  serv_out_time_s=%d\n",cli_out_time_s,serv_out_time_s);
	             printf("cli_out_time_us=%d  serv_out_time_us=%d\n\n",cli_out_time_us,serv_out_time_us);
	             printf("tam a zpet/2= %f %lf %lf\n",casPom,serv_out_time_s,serv_out_time_us);	
#endif	
	             if (casPom < casMin ) {casMin = casPom;}
	             if (casPom > casMax ) {casMax = casPom;}
	
	             casSum += casPom;
                     fprintf(outf,"%d. doba cesty:  %2.6f ms c_cas=%2.6f rychlost:  %2.3f Mbps<BR>\n",i,casPom*1000,casSum,(param->length/casPom)*8/Mbps);
	
	             break;
	} //ennd switch
    } //end for
#ifdef DEBUG2	
    printf("***********************   odeslano:%d paketu\n",i);
#endif	
    casAvg = casSum / (float)param->count;
    SumLength=(float)param->length*(float)param->count*8/Mbps;      //prepocet na Mbits
    printf("SumLength=:%f casSum=%f\n",SumLength,casSum);
    printf("UDP konec\n");
    /*kvuli statistice,kdyz se strati vse*/
    if (lost==param->count) {
       fprintf(outf, UDP_STATISTICS_SHORT,
	             param->count,
	             param->length,
             	     lost,(float)lost*100/(float)param->count);
       return 1;
		

    }
    fprintf(outf ,UDP_STATISTICS,
	    param->count,
	    param->length,
	    param->count*param->length,
	    lost,(float)lost*100/(float)param->count,
	    casSum,
	    SumLength/casSum,
	    casMin, casAvg, casMax,
	    param->length/casMax*8/Mbps, param->length/casAvg*8/Mbps, param->length/casMin*8/Mbps);
    free(buffer);
    return 1;
}
#endif

