/*
 * Autor: Martin Kot
 * Funkce generovani odpovedi na HTTP pozadavky
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <unistd.h>
#include "parser.h"
#include "init.h"
#include "util.h"
#include "confs.h"
#include "mutils.h"
#include "dynamAnsfer.h"
#include "pageStrings.h"

//HTTP stavove kody
char *codes[]={"200 OK",
"201 Created",
"400 Bad Request",
"403 Forbidden",
"408 Request Timeout",
"404 Not Found",
"501 Not Implemented",
"503 Service Unavailable"};

char **codesInHTML=codes;
char *errMessages[]={"OK",
"Created",
"Bad request",
"You don't have permission to access this URL.",
"Request timeout",
"The requested URL was not found on this server.",
"This type of request is not implemented on this server.",
"This service is now unavailable."};

char *dayOfWeek[]={"Sun","Mon",  "Tue", "Wed", "Thu", "Fri", "Sat"};
char *monthOfYear[]={"Jan",  "Feb",  "Mar",  "Apr","May","Jun","Jul",  "Aug",  
                   "Sep",  "Oct",  "Nov","Dec"};

request defRequest = {GET,"\\","","text/html",1,1,{NULL,0}};

//kopirovani mezi soubory
int fdToFdCopy(FILE *destFile, FILE *sourceFile){
  char c;
  while((c=fgetc(sourceFile))!=EOF){
    fputc(c,destFile);
  }
  return 0;
}

/*
 * Zapise do souboru socketFile hlavicku HTTP odpovedi 
 * @param req Struktura s rozdelenym pozadavkem
 * @param code Index v poli navratovych kodu
 * @param contLen Je-li nenulovy, vypise se do hlavicky Content-Length s touto
 *	hodnotou
 * @param ost Retezec s dalsimi volbami vypsanymi primo do hlavicky
 */
void writeHeader(FILE *socketFile,request *req, int code,int contLen,char *ost){
  char buffer[OPT_MAX_LEN];
  struct tm *cas;
  time_t cas2;
  cas2=time(NULL);
  cas = gmtime(&cas2);

/*Definovani HTTP_VER_AS_REQUEST zpusobi, ze v odpovedi bude stejna verze
  HTTP protokolu jako v pozadavku*/
#ifdef HTTP_VER_AS_REQUEST
  sprintf(buffer,"%s/%i.%i %s \r\n",HTTP_PROTOCOL,req->major,req->minor,
#else
  sprintf(buffer,"HTTP/1.1 %s \r\n",
#endif
    codes[code]);
  logRequest(req,codes[code]);
  fwrite(buffer,strlen(buffer),1,socketFile);
  sprintf(buffer,"Date: %s, %02d %s %d %02d:%02d:%02d GMT\r\n"
    "Server: %s\r\n",dayOfWeek[cas->tm_wday],
    cas->tm_mday,monthOfYear[cas->tm_mon],cas->tm_year+1900,cas->tm_hour,
    cas->tm_min,cas->tm_sec,SERVER_NAME);  
  if(req->mimeType!=NULL){
    sprintf(buffer+strlen(buffer),"Content-Type: %s\r\n",req->mimeType);
  }
  if(contLen!=0){
    sprintf(buffer+strlen(buffer),"Content-length: %d\r\n",contLen);
  }
  if(ost!=NULL){
    sprintf(buffer+strlen(buffer),"%s",ost);
  }
  sprintf(buffer+strlen(buffer),"\r\n");
  fwrite(buffer,strlen(buffer),1,socketFile);
}

/*
 * Posle odpoved na chybny pozadavek
 * @param errCode Chybovy kod
 */
void errorRequest(FILE *socketFile,HTTP_ERR_CODES errCode,request *req){
  struct utsname u;
  char buffer[OPT_MAX_LEN];
  uname(&u);
  req->mimeType="text/html";
  sprintf(buffer,"<HTML><HEAD><TITLE>%s</TITLE>\n<meta http-equiv=\""
    "Content-Type\" content=\"text/html; charset=ISO 8859-1\"></HEAD>\n"
    "<BODY><H1>%s</H1>%s<P>\n<HR><ADDRESS>"
    "%s Server at %s Port %d</ADDRESS></BODY></HTML>\r\n",
    codes[errCode],codesInHTML[errCode],errMessages[errCode],
    SERVER_NAME, u.nodename,HTTP_SERV_PORT);
  writeHeader(socketFile,req, errCode,0,NULL);
  if(req->type==GET){
    fwrite(buffer,strlen(buffer),1,socketFile);
  }
}

/*
 * Rozpozna typ souboru pozadovaneho v HTTP pozadavku a pripadne doplni cestu
 */
int parsePath(request *req){
  char *pos;
  char *path;
  struct stat fileInfo;
  if(req->type != POST){
    pos=strchr(req->path,'?');
    if(pos==NULL){
      req->params=NULL;
    }
    else{
      *pos='\0';
      req->params=(char *)malloc((strlen(pos+1)+1)*sizeof(char));
      strcpy(req->params,pos+1);
    }
  }
  pos=strrchr(req->path,'/');
  if(pos!=NULL){
    if(strcmp(pos+1,OUT_FILE)==0){
      return PROGRAM_OUTPUT;
    }
  }
  path = (char *)malloc((strlen(DOCUMENT_ROOT)+strlen(req->path)+1)*sizeof(char));
  strcpy(path,DOCUMENT_ROOT);
  strcat(path,req->path);
  free(req->path);
  req->path=path;
  if(stat(req->path,&fileInfo)==-1){
    return NO_EXIST;
  }
  if(S_ISDIR(fileInfo.st_mode)){
    int l=strlen(req->path);
    if(req->path[l-1]!='/'){
      path=(char *)malloc((strlen(DEF_FILE)+l+2)*sizeof(char));
      strcpy(path,req->path);
      strcat(path,"/");
      strcat(path,DEF_FILE);
    }
    else{
      path = (char *)malloc((strlen(DEF_FILE)+l+1)*sizeof(char));
      strcpy(path,req->path);
      strcat(path,DEF_FILE);
    }
    free(req->path);
    req->path=path;
  } 
  if(strstr(req->path,"/../")!=NULL){
    return TWO_DOTS;
  }
  if(stat(req->path,&fileInfo)==-1){
    return NO_EXIST;
  }
/*#ifdef NO_LINKS
  if(!S_ISREG(fileInfo.st_mode)){
#else
  if(!S_ISREG(fileInfo.st_mode)&&!S_ISLNK(fileInfo.st_mode)){
#endif
    return NO_EXIST;
  }*/
  if(!S_ISREG(fileInfo.st_mode)){
    return NO_EXIST;
  }
  else{
    pos=strrchr(req->path,'.');
    if(pos==NULL){
      req->mimeType=extToType("");
    }
    else{
      req->mimeType=extToType(pos+1);
    }
    return PATH_OK;
  }
}

/*
 * Odpoved na pozadavky typu GET, POST a HEAD
 */
void get(FILE *socketFile,request *req){
  FILE *sourceFile;
  int retVal,len;
  struct stat fileInfo;
  if((retVal=parsePath(req))==NO_EXIST){
    errorRequest(socketFile,HTTP_NOT_FOUND,req);
    LOG_ERR_S("HTTP","Pozadovan neexistujici soubor", req->path);
    return;
  }
  if(retVal==TWO_DOTS){
    errorRequest(socketFile,HTTP_BAD_REQUEST,req);
    LOG_ERR_S("HTTP","Pozadovan soubor s /../ v ceste", req->path);
    return;
  }
  if(retVal==PROGRAM_OUTPUT){
    analyze(socketFile,req);
    return;
  }
    
  if((sourceFile=fopen(req->path,"r"))==NULL){ 
    errorRequest(socketFile,HTTP_NO_PERMISSION,req);
    LOG_ERR_S("HTTP","Pozadovan neexistujici soubor", req->path);
    return;
  }
  if(stat(req->path,&fileInfo)==-1){
    len=0;
  }
  else{
    len=fileInfo.st_size;
  }
  writeHeader(socketFile,req,HTTP_OK,len,NULL);
  if(req->type==GET){
    fdToFdCopy(socketFile,sourceFile);
  }
  fclose(sourceFile);
}

/*
 * Odpoved na HTTP pozadavek
 * @param bufferTCP Pozadavek tak, jak byl nacten ze socketu, delky length
 * @return -1, neni-li pozadavek kompletni
 */
int ansfer(FILE *socketFile,char *bufferTCP,int length){
  request *req;
  int i;
  req=malloc(sizeof(request));
  req->mimeType=NULL;
  req->params=NULL;
  req->type = ERROR;
  req->path=NULL;
  req->opts.opt=NULL;
  req->opts.count=0;
  req->major=0;
  req->minor=0;

  if(analyzuj(bufferTCP,length,req)<0){
    return -1;
  }
  switch(req->type){
    case POST:
    case HEAD:
    case GET: get(socketFile,req);
              break;
    case LINK:
    case UNLINK:
    case PUT:
    case DELETE:
              errorRequest(socketFile,HTTP_NOT_IMPLEMENTED,req); 
              break;
    default:  errorRequest(socketFile,HTTP_BAD_REQUEST,req);
              break;
  }
  if(req->path!=NULL){
    free(req->path);
  }
#ifdef PRINT_OPTS_VALS
  for(i=0;i<req->opts.count;i++){
     printf("%s - %s\n",
	req->opts.opt[i].name,req->opts.opt[i].value);
  }
#endif
  for(i=0;i<req->opts.count;i++){
    if(req->opts.opt[i].name!=NULL){
      free(req->opts.opt[i].name);
    }
    if(req->opts.opt[i].value!=NULL){
      free(req->opts.opt[i].value);
    }
  }
  if(req->opts.opt!=NULL){
    free(req->opts.opt);
  }
  if(req->params!=NULL){
    free(req->params);
  }
  
  return 1;
}

