/*
 * Autor: Martin Kot
 * Zakladni funkce HTTP serveru - odchyceni pozadavku, main, ...
 */
#define _GNU_SOURCE
#include <stdlib.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
#include <sys/utsname.h>


#include "parser.h"
#include "mime.h"
#include "confs.h"
#include "util.h"
#include "mutils.h"

extern request defRequest;
extern mimeTypes mts;
FILE *socketFile;
int pipe_signal;
#ifdef MULTIPROCES
  int childCount=0;
#endif

/*
 * Osetreni signalu:
 * SIGCHLD - ukonci syny ve stavu zombie
 * SIGPIPE - prestane posilat data do preruseneho socketu
 * SIGINT a SIGTERM - ukonci aplikaci
 */
void sighndl(int sig){
  #ifdef MULTIPROCES
    int retval;
    if(sig==SIGCHLD){
      while(waitpid(-1,&retval,WNOHANG)>0){
        childCount--;
      }
      signal(SIGCHLD,sighndl);
    }
  #endif
  if(sig==SIGPIPE){
    LOG_ERR_S("HTTP","Preruseno spojeni s klientem",HTTP_client_addr);
    if(freopen("/dev/null","r",socketFile)==NULL){
      LOG_ERR("HTTP","Nelze otevrit soubor /dev/null");
    }
    pipe_signal=1;
    kill(getpid(),SIGUSR2);
  }
  if(sig==SIGTERM){
    if(reqLogFile!=NULL){
      fclose(reqLogFile);
    }
    printf("Obdrzen SIGTERM, ukoncuji server...\n");
    LOG_ERR("HTTP","Obdrzen SIGTERM, ukoncuji server...");
    exit(0);
  }
  if(sig==SIGINT){
    if(reqLogFile!=NULL){
      fclose(reqLogFile);
    }
    printf("Obdrzen SIGINT, ukoncuji server...\n");
    LOG_ERR("HTTP","Obdrzen SIGINT, ukoncuji server...");
    exit(0);
  }
}

/*
 * Posloucha na portu a prijima pozadavky. Pro vyrizeni pozadavku se forkne
 */
int HTTPServer(void) {
  struct sockaddr_in serv_addr, cli_addr;
  struct timeval timeout;
  int budik;
  int clilen;
  int readLen; 
  int actReadLen; 
  int sel;
  int socketFD, newSocketFD;
  int socketBuffer=BUF_SOCKET;
  int  bufLenTCP = BUF_LEN_TCP;
  char *bufferTCP;  
  struct utsname u;
  int reuse_val=1;
  char *ip;
  #ifdef MULTIPROCES
    pid_t syn;
  #endif

  fd_set rfds; 

  if ((socketFD = socket(AF_INET, SOCK_STREAM, 0)) < 0) { 
     LOG_ERR("HTTP","Nemohu otevrit socket");
     return 1;
  }
  setsockopt(socketFD, SOL_SOCKET, SO_RCVBUF, &socketBuffer, 
    sizeof(socketBuffer));
  setsockopt(socketFD, SOL_SOCKET, SO_REUSEADDR, &reuse_val,sizeof(reuse_val));
	
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(HTTP_SERV_PORT);

  if (bind(socketFD, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
     LOG_ERR("HTTP","Nemohu provest 'bind'");
     return 1;
  }
  uname(&u);
  ip=all_to_IP(u.nodename);
  printf("\nPosloucham na http://%s:%d\n",u.nodename,HTTP_SERV_PORT);
  if(ip!=NULL){
    printf("       neboli http://%s:%d\n\n",ip,HTTP_SERV_PORT);
    free(ip);
  }
  readTypes(&mts);
  listen(socketFD, MAX_CONNECTIONS);  //poslech max. 5 spojeni (?)


  clilen = sizeof(cli_addr);
  // Pockej na spojeni od klientu a odpovidej.
  signal(SIGTERM,sighndl);    
  signal(SIGINT,sighndl);    
  signal(SIGUSR2,SIG_IGN);    
  #ifdef MULTIPROCES
    signal(SIGCHLD,sighndl);    
    signal(SIGPIPE,SIG_IGN);    
  #else
    signal(SIGPIPE,sighndl);    
  #endif
  for ( ; ; ) {
    readLen=0;
    actReadLen=0;
    #ifdef MULTIPROCES
      while(childCount>=CONF_MAX_CLIENTS);
    #endif
    newSocketFD = accept(socketFD, (struct sockaddr *)&cli_addr, 
       (socklen_t *)&clilen);
    if (newSocketFD < 0) {
      LOG_ERR("HTTP","Chyba pri 'accept'");
      continue; 
    }
    #ifdef MULTIPROCES
      syn=fork();
      if(syn<0){
        close(newSocketFD);
        LOG_ERR("HTTP","Chyba pri fork");
        continue;
      }
      if(syn>0){
        close(newSocketFD);
        childCount++;
        continue;
      }
      close(socketFD);
      signal(SIGPIPE,sighndl);
      signal(SIGTERM,SIG_DFL);    
      signal(SIGINT,SIG_DFL);    
    #endif
    HTTP_client_addr=inet_ntoa(cli_addr.sin_addr);
    HTTP_client_port=cli_addr.sin_port;

    pipe_signal=0;
    bufferTCP = (char *) malloc((bufLenTCP+1) * sizeof(char));
        
    FD_ZERO(&rfds);
    timeout.tv_sec=CONF_TIMEOUT;
    timeout.tv_usec=0;
	    
    FD_SET(newSocketFD,&rfds);
    for ( ; ; ) {
      budik=0;
      for(;;){
        timeout.tv_sec=CONF_TIMEOUT;
        timeout.tv_usec=0;
	if(pipe_signal==1){
          break;
        }  
        sel=select(newSocketFD+1,&rfds,NULL,NULL,&timeout);
        if (sel==-1) {
          LOG_ERR("HTTP","Chyba v selektu");
        }
        if (sel>0) {
          break;
        }
        if(sel==0){
          budik=1;
          break;
        }
      } //for pro select
      if(pipe_signal==1){
        break;
      }  
      if(budik==1){
        if((socketFile=fdopen(newSocketFD,"r+"))==NULL){
          LOG_ERR("HTTP","Nelze otevrit socket jako FILE");
          break;
        }
        LOG_ERR_S("HTTP","Timeout pri prebirani pozadavku od",
           HTTP_client_addr);
        errorRequest(socketFile,HTTP_REQUEST_TIMEOUT,&defRequest);
        fflush(socketFile);
        fclose(socketFile);
        break;
      }
      actReadLen = read(newSocketFD, bufferTCP+readLen, 
                         bufLenTCP-readLen); //cteni dat
      if(pipe_signal==1){
        break;
      }  
      readLen+=actReadLen;
      if(bufLenTCP-readLen<=MIN_FREE){
        if((socketFile=fdopen(newSocketFD,"r+"))==NULL){
          LOG_ERR("HTTP","Nelze otevrit socket jako FILE");
          break;
        }
        errorRequest(socketFile,HTTP_BAD_REQUEST,&defRequest);
        fflush(socketFile);
        fclose(socketFile);
        break; 
      }
      *(bufferTCP + readLen) = '\0'; //pridej zarazku, aby slo vytisknou data
      if(pipe_signal==1){
        break;
      }  
      if((socketFile=fdopen(newSocketFD,"r+"))==NULL){
        LOG_ERR("HTTP","Nelze otevrit socket jako FILE");
        break;
      }
      if(ansfer(socketFile,bufferTCP,readLen)>0 || actReadLen == 0){
        signal(SIGUSR2,sighndl);    
        fflush(socketFile);
        fclose(socketFile);
        break;
      }
    } //for pro zpracovani pozadavku
    free(bufferTCP);

    #ifdef MULTIPROCES
      return 1;
    #endif
  } // for pro vsechny pozadavky
  return 0;
}

/*
 * Strucna napoveda
 */
void printUsage(char *name){
  printf("Program slouzi jako client pro testovani site a zaroven jako HTTP server pro\n");
  printf("ovladani tohoto testovani prostrednictvim HTML stranek.\n");
  printf("Uziti: %s [-d serverroot] [-f config] [-h]\n",basename(name));
  printf("serverroot   Nastavi pocatecni hodnotu ServerRoot directivy na serverroot.\n");
  printf("             Ta muze byt zmenena prikazem ServerRoot v konfiguracnim souboru.\n");
  printf("             Defaultni hodnota je %s\n",DEFAULT_SERVER_ROOT);
  printf("config       Vykona prikazy v souboru config pri startu. Pokud config nezacina\n");
  printf("             znakem /, je cesta brana relativne k ServerRoot. Defaultni\n");
  printf("             hodnota je %s\n",DEFAULT_CONFIG_FILE);
  printf("-h           Vypise tuto napovedu\n");
}


/*
 * Spousteci funkce
 * Nacte konfiguraci, preotvira stderr, otvira log pristupu, 
 * spousti HTTPServer(...)
 */
int main(int argc, char **argv) {
  char c;
  char *celeJmeno; 
  confServerRoot=NULL;
  confConfigFile=NULL;
  reqLogFile=NULL;
  while ((c=getopt(argc, argv, HTTP_S_PARAMS)) >= 0) {
    switch (c) {
      case 'd' : /* Adresar, kde budou hledany konfiguracni soubory */
                 confServerRoot=optarg;
                 break;
      case 'f' : /* Adresar, kde budou hledany konfiguracni soubory */
                 confConfigFile=optarg;
                 break;
      case 'h' : /* Napoveda */
                 printUsage(argv[0]);
                 exit(0);
                 break;
    }
  }

  setConfs(confServerRoot,confConfigFile);

  if(ERROR_LOG_FILE!=NULL){
    if(*ERROR_LOG_FILE=='/'){
      celeJmeno=strdup(ERROR_LOG_FILE);
      if(celeJmeno==NULL){
        LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
        return 1;
      }
    }
    else{
      celeJmeno=malloc((strlen(ERROR_LOG_FILE)+strlen(SERVER_ROOT)+2)
  		*sizeof(char));
      if(celeJmeno==NULL){
        LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
        return 1;
      }
      sprintf(celeJmeno,"%s/%s",SERVER_ROOT,ERROR_LOG_FILE);
    }
    printf("Menim chybovy vystup na: %s\n",celeJmeno);
    if(freopen(celeJmeno,"a",stderr)==NULL){
      fprintf(stdout, "Nelze zmenit chybovy vystup na %s\n",celeJmeno);
      exit(0);
    }
    free(celeJmeno);
  }

  if(CUSTOM_LOG_FILE!=NULL){
    if(*CUSTOM_LOG_FILE=='/'){
      celeJmeno=strdup(CUSTOM_LOG_FILE);
      if(celeJmeno==NULL){
        LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
      }
    }
    else{
      celeJmeno=malloc((strlen(CUSTOM_LOG_FILE)+strlen(SERVER_ROOT)+2)
  		*sizeof(char));
      if(celeJmeno==NULL){
        LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
      }
      else{
        sprintf(celeJmeno,"%s/%s",SERVER_ROOT,CUSTOM_LOG_FILE);
      }
    }
    if(celeJmeno!=NULL){
      LOG_ERR_S("HTTP-info","Oteviram logovaci soubor",celeJmeno);
      if((reqLogFile=fopen(celeJmeno,"a"))==NULL){
        LOG_ERR_S("HTTP", "Nelze otevrit logovaci soubor",celeJmeno);
        reqLogFile=NULL;
      }
      free(celeJmeno);
    }
    else{
      reqLogFile=NULL;
    }
  }
  HTTPServer();
  return 0;
}
	
