/*
 * Autor: Martin Kot
 * Pomocne funkce
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <time.h>
#include "util.h"
#include "parser.h"
#include "confs.h"
#define DELKA_ADRESY 100

extern const char *commands[];
char *HTTP_client_addr;
int HTTP_client_port;
FILE *reqLogFile;

/*
 * Vypise datum a cas
 */
void printTime(){
  time_t cas;
  struct tm *c;
  time(&cas);
  c=localtime(&cas);
  fprintf(stderr,"[%.2d.%.2d.%2d %.2d:%.2d:%.2d]",
      c->tm_mday,c->tm_mon+1,c->tm_year+1900,c->tm_hour,c->tm_min,c->tm_sec);
}

/*
 * Vypise chybovou hlasku, jde-li do souboru, je vypsan i cas
 */
void logErr(const char *kde, const char *co){
  if(ERROR_LOG_FILE!=NULL){
    printTime();
    fprintf(stderr," - [%s]\n\t%s\n",kde,co);
  }
  else{
    fprintf(stderr,"[%s] %s\n",kde, co);
  }
  fflush(stderr);
}

/*
 * Vypise chybovou hlasku, jde-li do souboru, je vypsan i cas
 */
void logErrStr(const char *kde, const char *co, const char *param){
  if(ERROR_LOG_FILE!=NULL){
    printTime();
    fprintf(stderr," - [%s]\n\t%s: %s\n",kde,co,param);
  }
  else{
    fprintf(stderr,"[%s] %s: %s\n",kde, co, param);
  }
  fflush(stderr);
}

/*
 * Vypise chybovou hlasku, jde-li do souboru, je vypsan i cas
 */
void logErrInt(const char *kde, const char *co, const int param){
  if(ERROR_LOG_FILE!=NULL){
    printTime();
    fprintf(stderr," - [%s]\n\t%s: %d\n",kde,co,param);
  }
  else{
    fprintf(stderr,"[%s] %s: %d\n",kde, co, param);
  }
  fflush(stderr);
}

/*
 * Zapise do log souboru HTTP pozadavek
 */
void logRequest(request *req,char *ret){
  time_t cas;
  struct tm *c;
  time(&cas);
  c=localtime(&cas);
  if(reqLogFile!=NULL){
    fprintf(reqLogFile,"[%s] - [%.2d.%.2d.%2d %.2d:%.2d:%.2d] \n\t%s %s HTTP/%d.%d\n",
	HTTP_client_addr,
	c->tm_mday,c->tm_mon+1,c->tm_year+1900,c->tm_hour,c->tm_min,c->tm_sec,
	commands[req->type],req->path,req->major,req->minor);
    if(req->type==POST && req->params!=NULL){
      fprintf(reqLogFile,"\t%s\n",req->params);
    }
    fprintf(reqLogFile,"\t%s\n",ret);
    fflush(reqLogFile);
  }
}

/*
 * Vraci IP adresu 
 * @param IP Je retezec, obsahujici bud domenove jmeno nebo jiz IP adresu
 */
char *all_to_IP (char *IP){
  struct hostent *host;
  char buf[DELKA_ADRESY];
  host = gethostbyname(IP);
  if (host == NULL) {
    LOG_ERR_P("HTTP","Nepodarilo se ziskat IP adresu");
    return NULL;
  }else {
    sprintf(buf,"%d.%d.%d.%d",(unsigned char)host->h_addr[0],
                              (unsigned char)host->h_addr[1],
                              (unsigned char)host->h_addr[2],
                              (unsigned char)host->h_addr[3]);
    return strdup(buf);
  }
}

/**
 * Vytvori buffer s nahodnymi daty
 * @param buf  Vyplnovany buffer
 * @param len  Delka/velikost bufferu
 * @param data Data, kterymi se bude vyplnovat buffer (pokud neni NULL)
 */
void fillBuffer(char * buffer, int buffLen, const char * pattern) {
  int plen;
  unsigned char j;
  int i;
  char * pat = hexa2bytes(pattern, &plen);

  if (pat == NULL || plen == 0) {
    for(i=0, j=0; i < buffLen; i++, j++) {
      buffer[i]=j;
    }
  } else {
    for (i=0; i < buffLen; i++) {
      buffer[i] = pat[i % plen];
    }
    free(pat);
  }
}
