/*
 * Autor: Martin Kot
 * Funkce pro rozdeleni HTTP pozadavku
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "parser.h"
#include "dynamAnsfer.h"
#include "confs.h"
#include "util.h"
const char *commands[] = {"GET","PUT","POST","HEAD","DELETE","LINK","UNLINK"};

/*
 * Urci typ pozadavku
 */
int command(char *buffer,int length, request *req){
  int i=0;
  int j=0;
  int k=0;
  int prvku;
  char commBuf[COMM_MAX_LEN];
  for(;buffer[i]<=' ';i++){
    if(i==length){
      req->type=ERROR;
      return -i;
    }
  }
  while(buffer[i]!=' '){
    if(j==COMM_MAX_LEN-1){
      req->type=ERROR;
      return i;
    }
    if(i==length){
      req->type=ERROR;
      return -i;
    }
    commBuf[j]=buffer[i];
    i++;
    j++;
  }
  commBuf[j]='\0';
  prvku=sizeof(commands)/sizeof(char*);
  for(k=0;k<prvku;k++){
    if(strcmp(commBuf,commands[k])==0){
      req->type=k;
      return i;
    }
  }
  req->type=ERROR;
  return i;
}

/*
 * Urci cestu k pozadovanemu souboru
 */
int path(char *buffer,int length, request *req, int pos){
  int i=pos;
  int j=0;
  int buflen;
  char* pathBuf;
  pathBuf=(char *)malloc(BUFF_LEN*sizeof(char));
  if(pathBuf==NULL){
    LOG_ERR("HTTP", "Nepodarilo se alokovat pamet.");
    exit(0);
  }
  buflen = BUFF_LEN;
  for(;buffer[i]<=' ';i++){
    if(i==length){
      req->path=NULL;
      return -i;
    }
  }
  while(buffer[i]>' '){
    if(j>=buflen-2){
      buflen*=2;
      realloc(pathBuf,buflen*sizeof(char));
      if(pathBuf==NULL){
        LOG_ERR("HTTP", "Nepodarilo se alokovat pamet.");
        exit(0);
      }
    }
    pathBuf[j]=buffer[i];
    i++;
    j++;
  }
  pathBuf[j]='\0';
  req->path=(char *)malloc((strlen(pathBuf)+1)*sizeof(char));
  if(req->path==NULL){
    LOG_ERR("HTTP", "Nepodarilo se alokovat pamet.");
    exit(0);
  }
  strcpy(req->path,pathBuf);
  free(pathBuf);
  return i;
}

/*
 * Urci verzi HTTP
 */
int version(char *buffer,int length, request *req, int pos){
  int i=pos;
  int j=0;
  char numBuf[SHORT_BUFF]; 
  for(;buffer[i]<=' ';i++){
    if(i==length){
      req->minor=FIRST_VERS_MINOR;
      req->major=FIRST_VERS_MAJOR;
      return -i;
    }
    if(buffer[i]=='\r' && buffer[i+1]=='\n'){
      req->minor=FIRST_VERS_MINOR;
      req->major=FIRST_VERS_MAJOR;
      return i+2;
    }
  }
  req->minor=0;
  req->major=0;
  for(;buffer[i]!='/';i++){
    if(i==length){
      req->type=ERROR;
      return -i;
    }
    if((buffer[i]=='\r' && buffer[i+1]=='\n')||j==SHORT_BUFF-1){
      req->type=ERROR;
      return i;
    }
    numBuf[j]=buffer[i];
    j++;
  }
  i++;
  numBuf[j]='\0';
  if(strcmp(numBuf, HTTP_PROTOCOL)!=0){
    req->type=ERROR;
    return i;
  }
  j=0;
  for(;buffer[i]>='0' && buffer[i]<='9';i++){
    numBuf[j++]=buffer[i];
    if(i==length){
      req->type=ERROR;
      return -i;
    }
    if((buffer[i]=='\r' && buffer[i+1]=='\n')||j==SHORT_BUFF-1){
      req->type=ERROR;
      return i;
    }
  }
  numBuf[j]='\0';
  req->major=atoi(numBuf);
  j=0;
  if(i==length){
    req->type=ERROR;
    return -i;
  }
  if(buffer[i]!='.'){
    req->minor=0;
    req->major=0;
    req->type=ERROR;
    return i;
  }
  i++;
  for(;buffer[i]>='0' && buffer[i]<='9';i++){
    numBuf[j++]=buffer[i];
    if(i==length){
      req->type=ERROR;
      return -i;
    }
    if((buffer[i]=='\r' && buffer[i+1]=='\n')||j==SHORT_BUFF-1){
      req->type=ERROR;
      return i;
    }
  }
  numBuf[j]='\0';
  req->minor=atoi(numBuf);
  for(;buffer[i]!='\r';i++){
    if(i==length){
      return -i;
    }
  }
  if(buffer[i+1]=='\n'){
    return i+2;
  }
  return -i;
}

/*
 * Nacte volby v hlavicce pozadavku
 */
int optParser(char *buffer,int length, request *req, int pos){
  option *opt2;
  int k;
  int i=pos;
  int count,j;
  char *nameBuff;
  char *valBuff;
  char optBuff[OPT_MAX_LEN];
  req->opts.count=0;
  req->opts.opt=NULL;
  
  while(1){
    if(buffer[i]=='\r' && buffer[i+1]=='\n'){
      return i+2;
    }
    for(j=0;buffer[i]!=':';i++){
      if(i==length){
        return -1;
      }
      if(buffer[i]=='\r' && buffer[i+1]=='\n'){
        break;
      }
      if(j<OPT_MAX_LEN-1){
        optBuff[j++]=buffer[i];
      }
    }
    if(buffer[i]!=':'){
      i+=2;
      continue;
    }
    optBuff[j]='\0';
    nameBuff=(char *)malloc((strlen(optBuff)+1)*sizeof(char));
    strcpy(nameBuff,optBuff);
    i++;
    for(;buffer[i]<=' ';i++){
      if(i==length){
        return -1;
      }
      if(buffer[i]=='\r' && buffer[i+1]=='\n'){
        i+=2;
        break;
      }
    }
    if(buffer[i]=='\r' && buffer[i+1]=='\n'){
      i+=2;
      continue;
    }
    for(j=0;;i++){
      if(i==length){
        return -1;
      }
      if(buffer[i]=='\r' && buffer[i+1]=='\n'){
        i+=2;
        optBuff[j]='\0';
        break;
      }
      if(j<OPT_MAX_LEN-1){
        optBuff[j++]=buffer[i];
      }
    }
    optBuff[j]='\0';
    valBuff=(char *)malloc((strlen(optBuff)+1)*sizeof(char));
    strcpy(valBuff,optBuff);
    count=req->opts.count;
    opt2 = (option *)malloc((2+count)*sizeof(option));
    for(k=0;k<count;k++){
      opt2[k]=req->opts.opt[k];
    }
    opt2[k].name=nameBuff;
    opt2[k].value=valBuff;
    free(req->opts.opt);
    req->opts.opt=opt2;
    req->opts.count++;
  }
  return i;
} 

/*
 * Spusti analyzu HTTP pozadavku
 */
int analyzuj(char *bufferTCP,int length, request *req){
  int pos;
  char *val;
/*  req->mimeType=NULL;
  req->params=NULL;
  req->type = ERROR;
  req->path=NULL;
  req->opts.opt=NULL;
  req->opts.count=0;
  req->major=0;
  req->minor=0;*/
  req->params=NULL;
  pos=command(bufferTCP,length,req);
  if(pos<0){
    return -1;
  }
  if(req->type == ERROR){
    return 0;
  }
  pos=path(bufferTCP,length,req,pos);
  if(pos<0){
    return -1;
  }
  if(req->path==NULL){
    return 0;
  }
  pos=version(bufferTCP,length,req,pos);
  if(pos<0){
    return -1;
  }
  if(req->minor==0 && req->major==0){
    return 0;
  }
  pos=optParser(bufferTCP,length,req,pos);
  if(pos<0){
    return -1;
  }
  if(req->type==POST){
    val=parCaseValue(&(req->opts),"Content-Length");
    if(val!=NULL){
      if((length-pos)<atoi(val)){
        return -1;
      }
    }
    req->params=malloc((strlen(bufferTCP+pos)+1)*sizeof(char));
    strcpy(req->params,bufferTCP+pos);
  }
  return 1;
}
