#include "init.h"
#include "util.h"
#include "icmp_util.h"
#include <errno.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <netinet/ip.h>

#define ICMP_D 4 /* Offset promenne casti ICMP zpravy */

extern int errno;

struct timeval * adj_t;
char do_send = TRUE; 		/* Kontrola, zda ma jeste posilat */
int buggy=TRUE;	              	/* Rozhozeny cas u gettimeofday a SIOCGSTAMP*/

/**
 * @param us - pocet mikrosekund, nez pokracuje
 * @returns chybu
*/ 
int ICMP_alarm_us(unsigned long long us) {
  struct itimerval iv;
  iv.it_interval.tv_sec = iv.it_value.tv_sec = us / 1000000;
  iv.it_interval.tv_usec = iv.it_value.tv_usec = (us % 1000000);
  return setitimer(ITIMER_REAL, &iv, NULL);
}

/**
 * @param cli_addr po provedeni zdrojova IP adresa 
 * @returns cislo soketu
*/ 
int ICMP_open(struct sockaddr_in * cli_addr) {
   int sockfdicmp, retval;
   socklen_t asize=sizeof(*cli_addr);
   char null_addr=FALSE;
   do_send = TRUE; 
   /* Reset adj_t - ocekavame, ze pri zmene delky bude soket znovu otevern */
   if (adj_t) { free(adj_t); adj_t = NULL; }	   

   if (cli_addr == NULL) { 
     null_addr=TRUE;
     cli_addr=(struct sockaddr_in*) malloc(asize);	 
   }
   
   cli_addr->sin_family      = AF_INET;
   cli_addr->sin_addr.s_addr = htonl(INADDR_ANY);
   cli_addr->sin_port        = htons(0); 

   if ((sockfdicmp = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) { 
       /*otevreni soketu pro ICMP*/
       LOG_ERR_P("ICMP", "Nemohu otevrit soket");
       retval = -1;
    } else { 
       retval=sockfdicmp;
       if (bind(sockfdicmp,(struct sockaddr *)cli_addr, asize) < 0) {
          LOG_ERR_P("ICMP", "Nemohu provest \'bind\'");
          retval = -1;
#ifdef IDEBUG   
       } else {
         LOG_ERR("ICMP", "Startuje v poradku");
#endif   
       }
    }	    
    if (null_addr) { free(cli_addr); } /*Pokud nebyla adresa parametrem - NULL*/
    return retval;
}

/**
 * @param buffer    pripraveny buffer 
 * @param buffLen   velikost bufferu
 */
void ICMP_echo_prep(char * buffer, int buffLen, const char * pattern) { 
  struct icmphdr * ih = (struct icmphdr*) (buffer);
  int plen;
  unsigned char j;
  int i,k;
  char * pat = hexa2bytes(pattern, &plen);

  ih->type=ICMP_ECHO;
  ih->code=0;
  ih->checksum = 0;

  if (pat == NULL || plen == 0) {
    for(i=ICMP_HDR_MIN + TS_LEN, j=0; i < buffLen; i++, j++) {
      buffer[i]=j;	  
    }	  
  } else {
    for (i=ICMP_HDR_MIN + TS_LEN, k=0; i < buffLen; i++, k++) {
      buffer[i] = pat[k % plen];
    }
    free(pat);
  }	  
}

/**
 * @param sock      soket
 * @param srv_addr  cilova IP adresa
 * @param buffer    posilana data
 * @param buffLen   delka posilanych dat
 * @param ttl       zivotnost odesilane ICMP zpravy
 * @param snd       cas odeslani ICMP zpravy
 */
int ICMP_echo_send(int sock, struct sockaddr_in * srv_addr, 
                   char * buffer, int buffLen, unsigned char ttl, 
                   struct timeval * snd)				{ 
  int result=0;
  struct icmphdr * ih = (struct icmphdr*) (buffer);
  struct timeval tv;
  int on = ttl;
  
  if (setsockopt(sock, SOL_IP, IP_TTL, &on, sizeof(on))) {
     LOG_ERR_P("ICMP", "Nelze nastavit IP_TTL");
     return -1;
  }

  do {
	  
    gettimeofday(&tv, NULL);
    *((struct timeval*) (buffer + ICMP_HDR_MIN)) = tv;
    ih->checksum=0;
    
    if (adj_t) { /* Uprava casu, aby kompenzovala zpozdeni */
      tv_sub(((struct timeval*) (buffer + ICMP_HDR_MIN)), adj_t);
    } else {
      adj_t = (struct timeval*) malloc(sizeof(struct timeval));

      gettimeofday(&tv, NULL);

      /* Zacatek mereneho useku */
      tv_sub(((struct timeval*) (buffer + ICMP_HDR_MIN)), adj_t);
      ih->checksum=inet_cksum((unsigned short *) ih, buffLen, 0);
      /* Konec mereneho useku */
      
      gettimeofday(adj_t, NULL);

      /* Rozdil pocatecni - koncovy, pri dalsim "odectu" pricteme */
      tv_sub(&tv, adj_t); 
      *adj_t = tv; 
      /* Za normalnich okolnosti je uprava kladna (bude odectena) */
      if (tv2usecs(adj_t) > 0) {
        LOG_ERR("ICMP", "timeofday bezi pozpatku, mereni nemusi byt presne!\n");
      }
#ifdef IDEBUG
      fprintf(stderr, "ICMP_echo_send: oprava = %s\n", 
		                                usecs2str(-tv2usecs(adj_t)));
#endif
      /* Aby continue proslo */
      result = -1;
      errno = EINTR; 
      continue;
    }
    
    ih->checksum=inet_cksum((unsigned short *) ih, buffLen, 0);
    if (do_send) {
      result=sendto(sock, buffer, buffLen, 0, 
		            (struct sockaddr*)srv_addr, sizeof(*srv_addr));
    } else {
      if (snd != NULL) { *snd = tv; }
      return -2;                              /* Koncime s odesilanim */
    }
  } while (result < 0 && (errno == EINTR ||   /* preruseno */
			  errno == ENOBUFS || /* nedostatek pameti pro sys. */
			  errno == ENOMEM ));	  
  if (result == -1) {
    result = -errno + SND_ERR_OFS;
    LOG_ERR_P("ICMP", "Odeslani echo request selhalo");
  }	  
  if (snd != NULL) { *snd = tv; }
  return result;
}

/**
 * @param sock      soket
 * @param srv_addr  zdrojova IP adresa
 * @param buffer    prijata data
 * @param buffLen   max. delka prijatych dat
 * @param sID	    ID ICMP toku (at neprijimame cizi ECHO REPLY) 
 * @param ttl       zivotnost prijate ICMP zpravy
 * @param rec       cas prijeti 
 */

int ICMP_echo_recv(int sock, struct sockaddr_in * srv_addr, 
                   char * buffer, int buffLen, unsigned char * ttl,
		   unsigned short sID, struct timeval * rec) {
  char * buff2; 
  struct iphdr * iph = (struct iphdr *) buffer; 
  struct icmphdr * ih = (struct icmphdr *) buffer; 
  int result, ip_hlen;
  int alen = sizeof(*srv_addr);
  
  do {
   result = recvfrom(sock, buffer, buffLen, 0, 
 		            (struct sockaddr *) srv_addr, &alen);
 
  } while (result < 0 && ( errno == EAGAIN ));

  if (result < 0) {
    if (errno == EINTR) { /* Preruseni signalem */
      return FAILED_INT; 
    } else { 
      LOG_ERR_P("ICMP", "Prijem ICMP zprav selhal");
    }
  } else {
    if (rec != NULL && ioctl(sock, SIOCGSTAMP, rec) < 0) {
      static int first = TRUE;
      register int err = errno;
      gettimeofday(rec, NULL);
      if (first) {
	errno = err;
	LOG_ERR_P("ICMP", "SIOCGSTAMP nelze pouzit, mereni bude mene presne");
	first = FALSE;
      }
    }	    

    /* Ulozeni adresy a ttl z IP hlavicky */
    srv_addr->sin_addr.s_addr = iph->saddr;
    *ttl = iph->ttl;

    /* Odstranit IP hlavicku a zmensit vystupni delku */
    ip_hlen = (iph->ihl) << 2;
    result -= ip_hlen;
    if (result < MIN_PING_SIZE) { return FAILED_LEN; }	/* Prilis kratke */
    buff2 = buffer + ip_hlen;
    memmove(buffer, buff2, result);
    
#if IDEBUG & 2
    fprintf(stderr,"HLEN v byte:%i: HLEN:%i, result: %i\n",
		    ip_hlen, iph->ihl, result
    );
#endif
    
    /* IP hlavicka odstranena - zjistit typ zpravy */
    if (ih->type == ICMP_ECHOREPLY) {   /* To nam ma prijit */
      if (ih->un.echo.id != sID) { /* ignorovane cislo streamu */
	 return OK_IGNORE;
      }	      
      if (inet_cksum((unsigned short *) ih, 
	             result, ~ih->checksum ) != ih->checksum) {
	 result = FAILED_CRC;
      }

#if IDEBUG & 2
      fprintf(stderr,"IH checksum: 0x%04x\n", ih->checksum); 
      fprintf(stderr,"IN checksum: 0x%04x\n", inet_cksum((unsigned short *) ih, result, ~ih->checksum)); 
#endif
      
    } else if (ih->type == ICMP_ECHO) { /* ignorujeme, (typ 0) */
      result = OK_IGNORE;
    } else {
      result = - ih->type; /* Vracime chybu se zap. cis. zpravy (-17 - -3)*/ 
    }	    
  }
      
#if EDEBUG & 4 /* Umyslne generovani CRC chyby pri prijmu */
      if ( rand() < ERR_TRESHOLD ) { result = FAILED_CRC; } 	  
#endif      
      
#if EDEBUG & 8 /* Umyslne generovani chybneho typu pri prijmu */
      if ( rand() < ERR_TRESHOLD/2 ) { result = - ICMP_DEST_UNREACH; }  
#endif      
      
  return result;
}

/**
 * @param buffer    prijata data
 */
char * ICMP_message(char * buffer) {
   static char sbuff[ICMP_STR_MAX+1];
   char * outb = sbuff;
   struct icmphdr * ih = (struct icmphdr*) (buffer);

   sbuff[0] = '\0'; 		/* Inicializace n azacatku */
   
   switch (ih->type) {
      case ICMP_ECHO:		strcat(outb, "Echo - pozadavek");
  	 break;

      case ICMP_ECHOREPLY:	strcat(outb, "Echo - odpoved");
	 break;

      case ICMP_DEST_UNREACH:	strcat(outb, "Nedosazitelny cil - ");
                   switch ( ih->code ) {
                     case ICMP_NET_UNREACH:  strcat(outb, "nedostupna sit");
                         break;
                     case ICMP_HOST_UNREACH: strcat(outb, "nedostupny uzel");
                         break;
                     case ICMP_PROT_UNREACH: strcat(outb, "nevhodny protokol");
                         break;
                     case ICMP_PORT_UNREACH: strcat(outb, "uzavreny port");
                         break;
                     case ICMP_FRAG_NEEDED: 
                         strcat(outb, "fragmentace nutna ale zakazana");
                         break;
                     case ICMP_SR_FAILED: 
                         strcat(outb, "selhalo explicitni smerovani");
                         break;
                     case ICMP_NET_UNKNOWN:   strcat(outb, "neznama sit");
                         break;
                     case ICMP_HOST_UNKNOWN:  strcat(outb, "neznamy uzel");
                         break;
                     case ICMP_HOST_ISOLATED: strcat(outb, "izolovany uzel");
                         break;
                     case ICMP_PKT_FILTERED:  strcat(outb,"odfiltrovany paket");
                         break;
                     case ICMP_PREC_VIOLATION:strcat(outb, "mimo poradi");
                         break;

                     default: strcat(outb, "neznama pricina");
                   }
	 break;

      case ICMP_SOURCE_QUENCH: 	strcat(outb, "Zpomalit zdroj"); 
	 break;

      case ICMP_REDIRECT:  	strcat(outb, "Presmerovan");
                   switch ( ih->code ) {
                     case ICMP_REDIR_NET:     strcat(outb, "a sit");
                             break;
                     case ICMP_REDIR_HOST:    strcat(outb, " uzel");
                             break;
                     case ICMP_REDIR_NETTOS:  strcat(outb,"a sit (typ sluzby)");
                             break;
                     case ICMP_REDIR_HOSTTOS: strcat(outb," uzel (typ sluzby)");
                             break;
                     default: strcat(outb, "o");
                   }
		   outb += strlen(outb); 
                   snprintf(outb, ICMP_STR_MAX - strlen(sbuff),
				 " pres %u.%u.%u.%u", buffer[ICMP_D],
                                                      buffer[ICMP_D+1],
                                                      buffer[ICMP_D+2],
                                                      buffer[ICMP_D+3]
		   );
	 break;

      case ICMP_TIME_EXCEEDED:	strcat(outb, "Cas vyprsel");
				if (ih->code == 1) {
				  strcat(outb, " (pri spojovani fragmentu)");
				}
	 break;

      case ICMP_PARAMETERPROB:	strcat(outb, "Chyba v hlavicce");
				if (ih->code == 0) {
		                  outb += strlen(outb); 
				  snprintf(outb, ICMP_STR_MAX - strlen(sbuff),
						 "Na pozici: %hu", 
                                                               buffer[ICMP_D]
				  );
             			}
	 break;

      case ICMP_TIMESTAMP: 	strcat(outb, "Casove razitko - pozadavek");
	 break;

      case ICMP_TIMESTAMPREPLY:	strcat(outb, "Casove razitko - odpoved");
	 break;

      case ICMP_INFO_REQUEST: 	strcat(outb, "Informace - pozadavek");
	 break;

      case ICMP_INFO_REPLY:	strcat(outb, "Informace - odpoved");
	 break;

      case ICMP_ADDRESS:	strcat(outb, "Sitova maska - pozadavek");
	 break;

      case ICMP_ADDRESSREPLY:	strcat(outb, "Sitova maska - odpoved");
	 break;
      default: snprintf(outb, ICMP_STR_MAX, "Neznamy typ zpravy (%i)",ih->type);
   }
   return sbuff;
}
