#ifndef __ICMP_UT_H
#define  __ICMP_UT_H
#include <netinet/ip_icmp.h>

#define ICMP_HDR_MIN 8 /* Min. delka ICMP hlavicky */
#define IP_HDR 20      /* Delka IP hlavicky        */
#define TS_LEN 8       /* Delka cas. znacky        */
#define MIN_PING_SIZE TS_LEN
#define MIN_ECHO (IP_HDR + ICMP_HDR_MIN + TS_LEN)

#define OK_IGNORE     0 	/* Ignorovane pakety    		*/
#define FAILED_CRC -100 	/* Nesedi kontrolni soucet  		*/
#define FAILED_INT -200 	/* Prijem prerusen signalem 		*/
#define FAILED_LEN -300 	/* Prilis kratky prijaty paket 		*/

#define SND_ERR_OFS -10000 	/* Posun errna do vice zap. cisel u 
                                   ICMP_echo_send pro rozliseni chyb    */

#define ICMP_STR_MAX 100	/* Max. delka retezce v ICMP_message	*/

/* Nastaveni alarmu s automatickym opakovanim po ms milisekundach */
#define ICMP_alarm(ms) ICMP_alarm_us(1000 * (ms))

/* Nastaveni alarmu s automatickym opakovanim po us mikrosekundach */
int ICMP_alarm_us(unsigned long long us);

/* Oteverni ICMP RAW soketu */
int ICMP_open( struct sockaddr_in * cli_addr);

/* Priparava ICMP zpravy pro odeslani */
void ICMP_echo_prep(char * buffer, int buffLen, const char * pattern);

/* Odeslani ICMP zpravy */
int ICMP_echo_send(int sock, struct sockaddr_in * srv_addr, 
                   char * buffer, int buffLen, unsigned char ttl,
                                             struct timeval * snd);

/* Prijem ICMP zpravy */
int ICMP_echo_recv(int sock, struct sockaddr_in * srv_addr, 
                   char * buffer, int buffLen, unsigned char * ttl,
		   unsigned short sID, struct timeval * rec);

/* Vypis informace z ICMP zpravy */
char * ICMP_message(char * buffer);

/* Kod z ICMP zpravy */
#define ICMP_code(data) (((struct icmphdr *) data)->code)

/* Typ z ICMP zpravy */
#define ICMP_type(data) (((struct icmphdr *) data)->type)

/* ID ICMP echo streamu */
#define ICMP_echo_ID(data) (((struct icmphdr *) data)->un.echo.id)

/* Sekvencni cislo v ICMP echo streamu */
#define ICMP_echo_seqNo(data) (((struct icmphdr *) data)->un.echo.sequence)

/* Nastaveni stavove informace, zda odesilat */
extern char do_send; 

#endif  /*__ICMP_UT_H*/
