#include "init.h"
#include "util.h"
#include "messages.h"
#include "icmp_util.h"
#include "ping.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <time.h>

/* Default parametry PINGu   */
PINGER def_png = {20, 1000, 56, 255, {0}, 0, 0 }; 

PING_STATE ping_state;   /* Kompletni stavova informace beziciho PINGu       */
char dups [DUP_WIN_SIZ]; /* Okynko pro kontrolu duplicit prichozich odpovedi */

/*--------------ping_finalize() signal handler ---------------------------*/
void ping_finalize(int sigNo) {
    int adj;
    ping_state.running = FALSE; /* Ukoncen dobeh */
    ICMP_alarm(0);
    signal(SIGALRM, SIG_IGN);
    adj = ping_state.ps->rec - ping_state.p->replies;
    if (adj > 0) { 
       /* Sice na ne pockame, ale do vysledku je nezahrneme pokud prijdou. 
	  Pokud se ovsem ztrati, pak se to v poctu ztracenych projevi 
       */
       ping_state.ps->rec -= adj;	    
       ping_state.ps->trans -= adj;	    
    }	    
}    

/*--------------ping_stop() signal handler ---------------------------*/
void ping_stop(int sigNo) {
    if (ping_state.ps->rec == ping_state.ps->trans) {
      ping_finalize(0);
    } else {
      ping_state.running = IGNORE; /* Pocitej zbyle prichozi */
      do_send = FALSE;
      signal(SIGALRM, ping_finalize);
      ICMP_alarm(1 + FINAL_MULT * ping_state.ps->rtt_max / 1000);
    }
    if (sigNo <= SND_ERR_OFS) { /* Odeslani selhalo z neznamych pricin */
      ping_state.running = sigNo;
    }				
}	

/*--------------ping_resend() signal handler -------------------------*/
void ping_resend(int sigNo) {
  struct icmphdr * ih = (struct icmphdr*) ping_state.data;
  int result;

/*  last_snd = TRUE; */
  if (sigNo == 0) {
    srand(time(NULL));
    ping_state.stream = ih->un.echo.id = rand();
    ih->un.echo.sequence = 0;
  } else {
    register unsigned short cnt = ++ih->un.echo.sequence;
    /*if (cnt == 0) { 
      ping_stop(0);
      return;
    }*/
  } 
  dups[ih->un.echo.sequence % DUP_WIN_SIZ] = 0;

  if ((result=ICMP_echo_send(ping_state.sock, ping_state.s_addr, 
		             ping_state.data, 
			     ping_state.p->pack_size + ICMP_HDR_MIN, 
                             ping_state.p->ttl, NULL)) < 0) {
      ping_stop(result);
      return; 
  }	  
  
#if EDEBUG & 2 /* Umyslne generovani duplikatu pri odeslani */
  if ( rand() < ERR_TRESHOLD ) {  
    if ((result=ICMP_echo_send(ping_state.sock, ping_state.s_addr, 
			       ping_state.data, 
			       ping_state.p->pack_size + ICMP_HDR_MIN, 
                               ping_state.p->ttl, NULL)) < 0) {
        ping_stop(result);
        return; 
    }
  }	  
#endif      
  
  ping_state.ps->trans++;
  signal(SIGALRM, ping_resend);
#if PDEBUG & 2
  fprintf(stderr, "Odeslana ICMP zprava %i\n", ping_state.ps->trans);
#endif		  
}	

/*-------------------------ping_help()--------------------------------*/
void ping_help() {
  fprintf(stderr, "Volby:");
  fprintf(stderr, 
     "\t-c pocet\tmin. odpovedi, po jejich prijeti program skonci\t[%i]\n"
     "\t-i ms\t\tcas mezi odeslanim pozadavku (ms)\t\t[%i]\n"
     "\t-s bajtu\tvelikost odeslaneho paketu\t\t\t[%i]\n"
     "\t-t ttl\t\tTime-to-live\t\t\t\t\t[%i]\n"
     "\t-v\t\tUkecany mod (vypise odpoved i ostatni ICMP zpravy)\n"
     "\t-q\t\tTichy mod\n"
     "\t-n\t\tJen IP adresa bez domenoveho jmena\n"
     "\t-p vzorek\tVzorek odesilanych dat (max. 16 bajtu)\t\t[%s]\n",
     def_png.replies, def_png.wait_ms, def_png.pack_size, (int) def_png.ttl,
     (def_png.pattern)?(def_png.pattern):"<Neni>");
  exit(-1);
}

/*-------------------------ping_info()--------------------------------*/
void ping_info(PINGER *p) {
  if ( p == NULL) {
    fprintf(stderr, "--> Zadne informace nejsou k dispozici! <--\n");
  } else {
    fprintf(stderr, "cnt=%i, wait=%i ms, size=%i, ttl=%i", p->replies, 
            p->wait_ms, p->pack_size, p->ttl
    );
    if (p->flags.quiet)  { fputs(", quiet", stderr); }
    if (p->flags.verbose)  { fputs(", verbose", stderr); }
    if (p->flags.numeric)  { fputs(", numeric", stderr); }
    fprintf(stderr, "\npattern: \'%s\'", (p->pattern)?(p->pattern):"<None>");
    fprintf(stderr, "\nhost: \'%s\'\n", (p->host)?(p->host):"<None>");
  };
}

/*-------------------------ping_setup()-------------------------------*/
PINGER * ping_setup(int opt_cnt, char **opts) {
   char c;
   PINGER * np = (PINGER *) malloc (sizeof(PINGER));
   *np = def_png;
   while ((c=getopt(opt_cnt, opts, PPARAMS)) >= 0) {
     switch (c) {
       case 'c' : /* Pocet prichozich odpovedi 	*/
                 np->replies=atoi(optarg);
                 break;
       case 'i' : /* Prodleva			*/
                 np->wait_ms=atoi(optarg);
                 break;
       case 's' : /* Velikost paketu 		*/
                 np->pack_size=(atoi(optarg) < 0xffff-MIN_ECHO) ? atoi(optarg)
			        : 0xffff - MIN_ECHO;
                 break;
       case 't' : /* Time-to-live 		*/
                 np->ttl=(atoi(optarg) < 256) ? atoi(optarg): 255;
                 break;
       case 'p' : /* Vzorek			*/
                 np->pattern=strdup(optarg); 
                 break;
       case 'n' : /* Jen IP adresa bez resolve	*/
                 np->flags.numeric=1;
                 break;
       case 'q' : /* Tichy mod 			*/
                 np->flags.quiet=1;
                 break;
       case 'v' : /* Ukecany mod 		*/
                 np->flags.verbose=1;
                 break;
       default:
                ping_help();          
     }
   }
   
   if (optind == opt_cnt - 1) {
     np->host = strdup(opts[optind]);
   } else {
     fprintf(stderr, 
             "%s: musi byt specifikovana prave 1 adresa ciloveho pocitace\n", 
	     opts[0]
     );
     ping_help();
   }

   if (ping_check_params(np) < 0) {
     ping_free(np, NULL);
     np = NULL;
     ping_help();
   }

   return np;
}

/*---------------------ping_check_params()----------------------------*/
int ping_check_params(PINGER *np) {
   if ((np->replies <= 0) || (np->pack_size < MIN_PING_SIZE) ||
       (np->wait_ms) <=0   || (np->ttl == 0) || (np->host == NULL)) {
#ifdef PDEBUG
     ping_info(np);
     printf( "np->replies <= 0  : %s\n",  (np->replies <= 0)?"Ano":"Ne");
     printf( "np->wait_ms <= 0   : %s\n",  (np->wait_ms <= 0)?"Ano":"Ne");
     printf( "np->pack_size < %i : %s\n", MIN_PING_SIZE,  
             (np->pack_size <= MIN_PING_SIZE)?"Ano":"Ne");
     printf( "np->ttl <= 0\t  : %s\n",  (np->ttl <= 0)?"Ano":"Ne");
     printf( "np->host == NULL  : %s\n",  (np->host == NULL)?"Ano":"Ne");
#endif
     return -1;
   }
   return 0;
}	

/*--------------------------ping_exec()-------------------------------*/
PING_STATS * ping_exec(PINGER * p, FILE * outf) {
  PING_STATS * ps = (PING_STATS *) calloc(sizeof(PING_STATS),1);
  int sock, i;    
  struct hostent * hostinf = NULL;                      /* Info od resolveru */
  struct sockaddr_in srv_adr, cli_adr;                  /* Adresa pro bind   */ 
  char * data = malloc(p->pack_size + MIN_ECHO);        /* Odchozi buffer    */
  char data_in [SOCK_BUFF];                             /* Prichozi buffer   */
  struct timeval ret_time;                              /* Cas prichodu      */
  unsigned char ttl;					/* Prichozi  TTL     */
  
  srv_adr.sin_family = AF_INET; 			/* Trida adres       */

  if (outf == NULL) {
    LOG_ERR("PING", "Vystup je zahazovan!");
    outf=fopen(DEV_NULL, "w+b");
  }
		  
  /** Vzdaleny: hostname, IP adresa */
  if (inet_aton(p->host, &srv_adr.sin_addr) != 1) {
    if ((hostinf = gethostbyname(p->host)) == 0) {
      LOG_ERR_P("PING", "Nelze zjistit IP adresu hostu");
      fprintf(outf, M_IP_NOTFOUND, p->host);
      fflush(outf);
      ping_free(NULL,ps);
      return NULL; 
    }
    free(p->host);
    p->host = strdup(hostinf->h_name);
    memcpy(&srv_adr.sin_addr, hostinf->h_addr_list[0], 4); /* Kopie adresy   */
  }
  sock=ICMP_open(&cli_adr);
  
  if (sock < 0) {  /* Chyba pri registraci */
    LOG_ERR("PING", "Nelze zacit (nedostatecna privilegia?)");
    fprintf(outf, M_EXEC_PR, "PING", p->host);
    fflush(outf);
    free(ps);
    return NULL;
  }	  
  fprintf(outf, M_PING_HEAD, p->host, 
                inet_ntoa(srv_adr.sin_addr), 
		p->pack_size, p->pack_size + IP_HDR + ICMP_HDR_MIN
  );
  fflush(outf);
  
  /* Aktivace PINGu */
  signal(SIGUSR2, ping_stop);
#ifdef PMAIN /* Na prani MK */ 
  signal(SIGINT, ping_stop);
  signal(SIGTERM, ping_stop);
#endif
  signal(SIGALRM, ping_resend);
  ping_state.running=TRUE;
  ping_state.ps=ps;
  ping_state.p=p;
  ping_state.data=data;
  ping_state.s_addr=&srv_adr;
  ping_state.sock=sock;
  
  /* Zakazat fragmentaci */
  i = IP_PMTUDISC_DO;
  if (setsockopt(sock, SOL_IP, IP_MTU_DISCOVER, &i, sizeof(i))) {
     LOG_ERR_P("ICMP", "Upozorneni: nelze nastavit DF");
  }
  
  /* Pripravime si v pameti paket */
  ICMP_echo_prep(data, p->pack_size + ICMP_HDR_MIN, p->pattern);
  /* 1. odeslany paket */
  ping_resend(0);
  ICMP_alarm(p->wait_ms);

  while (ping_state.running == TRUE) {
    int retval;
    retval = ICMP_echo_recv(sock, &srv_adr, data_in, SOCK_BUFF, 
		                  &ttl, ping_state.stream, &ret_time);
   
    /* Prijat signal -> break/casovac, mozna budeme koncit */
    if (retval == FAILED_INT || retval == OK_IGNORE) { continue;} 
    
    if (retval > 0) {
      ps->rec++;
      ping_info_update(outf,data_in, retval, &srv_adr.sin_addr, ttl, &ret_time);
    } else if (retval == FAILED_CRC) {	    
      ps->crcs++;
    } else {	    
      ps->errs++;
      if (ping_state.p->flags.verbose) { /* Vypis chyb jen v ukecanem rezimu */
        fprintf(outf, M_ICMP_MSG, ICMP_message(data_in));
        fflush(outf);
      }	
#ifdef PDEBUG	 
      fprintf(stderr, "Ping - vysledek ICMP_echo_recv(): %i\n", retval);
#endif
    }
    if (ps->rec >= p->replies) {
      ping_stop(0);
    }
  }

#ifdef M_PING_FTR
  fprintf(outf, M_PING_FTR);
  fflush(outf);
#endif  
  
  while (ping_state.running == IGNORE) { /* Ukoncovaci beh */
    int retval = ICMP_echo_recv(sock, &srv_adr, data_in, SOCK_BUFF, 
	  	                      &ttl, ping_state.stream, &ret_time);
    if (retval > 0) { ps->rec++; }	    
  }	  
  
  if (ping_state.running <= SND_ERR_OFS ) {
    fprintf(outf, M_ICMP_SND, strerror(-ping_state.running + SND_ERR_OFS));
    fflush(outf);
  }	  
  
  /* Deaktivace PINGu */
  ICMP_alarm(0);
  signal(SIGUSR2, SIG_DFL);
#ifdef PMAIN /* Na prani MK */ 
  signal(SIGINT, SIG_DFL);
  signal(SIGTERM, SIG_DFL);
#endif  
  signal(SIGALRM, SIG_IGN);
  ping_state.running=FALSE;
  ping_state.data=NULL;
  ping_state.s_addr=NULL;
  free(data);
  
  return ps;
}

/*----------------------ping_info_update()----------------------------*/
void ping_info_update(FILE * outf, char * buffer, int buffLen, 
		      struct in_addr *adr, char ttl, struct timeval *rec) {
  struct icmphdr * ih = (struct icmphdr*) buffer;
  struct timeval rtt;
  long long usec;
  if (ih->type == ICMP_ECHOREPLY)  { /* A to by melo byt vzdy! */
    timeval_subtract(&rtt, rec, (struct timeval *)(buffer+TS_LEN));
    usec = tv2usecs(&rtt);

    /* Kontrola zaporne delky intervalu */
    if (usec < 0) { usec = -usec; }

    /* Min. rtt */
    if (ping_state.ps->rtt_min == 0 || usec < ping_state.ps->rtt_min) {
      ping_state.ps->rtt_min = usec;
    }
    
    /* Max. rtt */
    if (usec > ping_state.ps->rtt_max) { ping_state.ps->rtt_max = usec; }
    
    /* Celkovy rtt */
    ping_state.ps->rtt_all += usec;
    
    /* Je-li zapnuty vypis */
    if (! ping_state.p->flags.quiet) {
      struct hostent * hostinf = NULL;             /* Info od resolveru */
      char * adr_IP = strdup(inet_ntoa(*adr));     /* Retezec s IP adresou */ 

      fprintf(outf, M_PINFO_BYTES, buffLen-ICMP_HDR_MIN);	    
      if (ping_state.p->flags.numeric || 
          !(hostinf = gethostbyaddr((char *)adr, 4, AF_INET)) ) {
        fprintf(outf, M_IP_ONLY, adr_IP);	    
      } else {
        fprintf(outf, M_IP_HOST, hostinf->h_name, adr_IP);	    
      }	      
      fprintf(outf, M_PINFO_INF, 
		    ih->un.echo.sequence, (unsigned char)ttl, usecs2str(usec), 
		    (dups[ih->un.echo.sequence % DUP_WIN_SIZ])?"(DUP!)":"" 
      );	    
      if (dups[ih->un.echo.sequence % DUP_WIN_SIZ] == 0) {
        dups[ih->un.echo.sequence % DUP_WIN_SIZ]=1;
      } else { /* Osetrime duplikat v celkovem poctu */
	ping_state.ps->dups++;
	ping_state.ps->rec--;
      }      
      free(adr_IP);
#ifdef PDEBUG
      if (ping_state.p->flags.verbose) { 
        hexshow(stderr, buffer, buffLen, FALSE);
      }	
#endif
    }	    
    fflush(outf);
  }	  
}	

/*-----------------------ping_rtt_stats()-----------------------------*/
void ping_rtt_stats(PINGER * p, PING_STATS * inf, FILE * outf) {
  
  if (outf == NULL) { /* Tak naco to volame? */
    LOG_ERR("PING", "Pozadavek zahazovat statistiky, proc tedy volano?");
    return; 
  } 
  if ( p != NULL) {
    fprintf(outf, M_RTT_STATS, p->host); 
  }
  if ( inf == NULL) {
    fprintf(outf, M_NOSTATS);
  } else if ( inf->trans ) {
     fprintf(outf, M_RTT_SNDREC, inf->trans, inf->rec );
     
     /* Duplicity */
     if (inf->dups != 0) { fprintf(outf, M_RTT_DUPS, inf->dups); }    
     
     /* Chybne CRC */
     if (inf->crcs != 0) { fprintf(outf, M_RTT_CRCS, inf->crcs); }     
     
     /* Chybove zpravy */
     if (inf->errs != 0) { fprintf(outf, M_RTT_ERRS, inf->errs); }     
     
     fprintf(outf, M_RTT_LOST,
                   ((inf->trans - inf->rec) * 100) / inf->trans
     );
     
     if ( inf->rec > 0 ) { /* Jen pokud neco prijmeme */
       fprintf(outf, M_RTT_RTTS, 
                     inf->rtt_min / 1000.0, 
		     inf->rtt_all / 1000.0 / (inf->rec + inf->dups),
                     inf->rtt_max / 1000.0
       );
     }
     
  } else {
    fprintf(outf, M_NOSEND);
  }
  putc('\n',outf); /* Odradkovani na konci */
  fflush(outf);
}

/*-------------------------ping_free()--------------------------------*/
void ping_free(PINGER * p, PING_STATS * inf) {
   assert(p || inf); 
   if (p) {
      if (p->pattern)
         free(p->pattern);
      if (p->host)
         free(p->host);
      free(p);
   }
   if (inf) {
      free(inf);
   }
} 

/*-------------------------------ping()-------------------------------*/
int ping(int opt_cnt, char **opts) {
   PINGER * p = ping_setup(opt_cnt, opts);
   PING_STATS * ps; 
#ifdef PDEBUG
   ping_info(p);
#endif
   ps = ping_exec(p, stdout);
   ping_rtt_stats(p, ps, stdout);
   ping_free(p, ps);
   return 0;
}

/*--------------------------ping_silent()-----------------------------*/
/* Ukazkova varianta bez vypisu - i kdyz jsou posilany do DEV_NULL    */
/*--------------------------------------------------------------------*/
int ping_silent(int opt_cnt, char **opts) {
   PINGER * p = ping_setup(opt_cnt, opts);
   PING_STATS * ps; 
   ps = ping_exec(p, NULL);
   /* Udelat neco s p a ps - parsovat vlastni funkci */
   ping_free(p, ps);
   return 0;
}

#ifdef PMAIN
int main(int argc, char ** argv) {
  return ping/*_silent*/(argc, argv);
}
#endif
