#ifndef __TRACERT_H
#define __TRACERT_H

#include "util.h"
#define TPARAMS "f:t:w:q:s:vnh"
#define MAX_TTLS 256

typedef struct {
  unsigned int verbose : 1, quiet : 1, numeric : 1;
  /*   ukecany mod^   tichy mod^    jen IP^  */
} tr_flag_t;

typedef struct {
  int queries;                /* Pocet pokusu pro kazdou hoodnotu TTL       */
  int wait_ms;                 /* Prodleva v sekundach do timeoutu           */
  unsigned short pack_size;   /* Delka ICMP hlavicky a dat                  */ 
  unsigned char min_ttl;      /* Pocatecni Time-To-Live pro ICMP zpravy     */
  unsigned char max_ttl;      /* Koncovy   Time-To-Live pro ICMP zpravy     */
  tr_flag_t flags;            /* Flagy                                      */ 
  char * host;                /* Cilovy pocitac, ktery pingame              */
} TRACER;

struct TRACER_STAT;
typedef struct {
  char * via;		      /* Adresa routeru, ppres ktery skaceme        */
  long long * times;          /* Pole: Casy prichodu (zaporny cas -> chyba) */
} TRACER_STAT;

typedef TRACER_STAT * TRACER_STATS;     /* Pole vysledku                   */

typedef struct {
  TRACER * t;				/* Konfigurace                     */
  TRACER_STATS * ts;			/* Vysledky                        */
  int sock;                             /* Soket, pres ktery posilame      */
  unsigned short stream;                /* Identifikace ICMP streamu       */
  int running;                          /* Stavova informace, zda bezi     */
} TRACER_STATE;

/* Kompletni tracert - jako main */
int tracert(int opt_cnt, char **opts);

/* Osetreni signalu -> ukonceni tracertu */
void tracert_stop(int sigNo);

/* Vypis napovedy */
void tracert_help();

/* Nastaveni parametru - primarne z prikazove radky */
TRACER * tracert_setup(int opt_cnt, char **opts);

/*Kontrola spravnosti rozsahu parametru */
int tracert_check_params(TRACER *t);

/* Informace o nastavenych volbach */
void tracert_info(TRACER *t);

/* Spusteni tracertu, vystup do streamu */
TRACER_STATS * tracert_exec(TRACER * t, FILE * outf);

/* Zpracovani 1 odpovedi */
void tracert_info_update(FILE * outf, int ttl, struct in_addr *adr, 
		         TRACER_STAT * ts, int q_count    ); 

/* Uvolni pamet drzenou v TRACER a TRACER_STATS */
void tracert_free(TRACER * p, TRACER_STATS * inf); 


#endif /*__TRACERT_H*/
