/*
 * Autor: Martin Kot
 * Funkce pro rozdeleni hodnot z formulare a volani prislusnych testovacich
 * funkci
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parser.h"
#include "init.h"
#include "ping.h"
#include "tracert.h"
#include "icmp_util.h"
#include "confs.h"
#include "mutils.h"
#include "pageStrings.h"
#include "dynamAnsfer.h"

extern PINGER def_png;
extern TRACER def_tr;
parametry def_tcpudp={10,1024,8192,SERV_PORT,NULL,DEF_DATA,100};

/*
 * Prevede specialni sekvence v pozadavku na bezne znaky
 */
char* escToChar(char *s){
  char *newS;
  char escSeq[3];
  int i;
  int j;
  if(s==NULL){
    return NULL;
  }
  newS=(char *)malloc((strlen(s)+1)*sizeof(char));
  if(newS==NULL){
    LOG_ERR("HTTP","Nedostatek pameti");
    exit(-1);
  }
  for(i=0,j=0;s[i]!='\0';i++){
    if(s[i]=='+'){
      newS[j++]=' ';
    }
    else if(s[i]=='%'){
      if(s[i+1]=='\0'){
        return NULL;
      }
      if(s[i+2]=='\0'){
        return NULL;
      }
      escSeq[0]=s[i+1];
      escSeq[1]=s[i+2];
      escSeq[2]='\0';
      sscanf(escSeq,"%x",(int *)&newS[j++]);
      i+=2;
    }
    else{
      newS[j++]=s[i];
    }
  }
  for(j--;newS[j]<=' '&&j>=0;j--); //preskoci white spaces na konci
  newS[++j]='\0';
  return newS;
}

/*
 * Rozdeli retezec hodnot z formulare na jednotlive
 */
options *parseParams(char *paramString){
  options *opts;
  option o;
  char *retVal;
  char *oneParam;
  char *allParams=strdup(paramString);
  
  opts=(options *)malloc(sizeof(options));
  opts->count=0;
  opts->opt=NULL;
  if(paramString==NULL){
    return opts;
  }
  do{
    retVal=strchr(allParams,'&');
    if(retVal==NULL){
      oneParam=allParams;
    }
    else{
      *retVal='\0';
      oneParam=allParams;
      allParams=retVal+1;
    }

    o.name=oneParam;
    o.value=strchr(oneParam,'=');
    if(o.value==NULL){
      o.value="";
    }
    else{
      *(o.value++)='\0';
    }
    o.value=escToChar(o.value);
    if(o.value!=NULL){
      if(strcmp(o.value,"")==0){
        o.value=NULL;
      }
    }
    o.name=escToChar(o.name);
    opts->opt=(option *)realloc(opts->opt,(opts->count+1)*sizeof(option));
    opts->opt[opts->count].name=o.name;
    opts->opt[opts->count].value=o.value;
    opts->count++;
  } while(retVal!=NULL);
  return opts;
}

/*
 * Vraci hodnotu parametru jmena parName - nerozlisuje velka a mala pismena
 */
char *parCaseValue(options *opts,char *parName){
  int i;
  for(i=0;i<opts->count;i++){
    if(strcasecmp(opts->opt[i].name,parName)==0){
      return opts->opt[i].value;
    }
  }
  return NULL;
}

/*
 * Vraci hodnotu parametru jmena parName - rozlisuje velka a mala pismena
 */
char *parValue(options *opts,char *parName){
  int i;
  for(i=0;i<opts->count;i++){
    if(strcmp(opts->opt[i].name,parName)==0){
      return opts->opt[i].value;
    }
  }
  return NULL;
}

/*
 * Porovna hodnotu parametru jmenem parName s hodnotou parValue
 */
int parValueCmp(options *opts,char *parName,char *parValue){
  int i;
  for(i=0;i<opts->count;i++){
    if(strcmp(opts->opt[i].name,parName)==0){
      if(strcmp(opts->opt[i].value,parValue)==0){
        return TRUE;
      }
      else{
        return FALSE;
      }
    }
  }
  return FALSE;
}
  
/*
 * Porovna hodnotu parametru jmenem parName s hodnotou parValue - nerozlisuje 
 * velka a mala pismena
 */
int parCaseValueCmp(options *opts,char *parName,char *parValue){
  int i;
  for(i=0;i<opts->count;i++){
    if(strcasecmp(opts->opt[i].name,parName)==0){
      if(strcasecmp(opts->opt[i].value,parValue)==0){
        return TRUE;
      }
      else{
        return FALSE;
      }
    }
  }
  return FALSE;
}

/*
 * Volani funkce pro ping a overeni hodnot z formulare
 */
int pingFunction(FILE *socketFile,options *opts){
  char buffer[OPT_MAX_LEN];
  char *value;
  PING_STATS *pstat;
  PINGER * np = (PINGER *) malloc (sizeof(PINGER));
  *np = def_png;
  np->host=parCaseValue(opts,PAGE_ADDRESS);
  if(np->host==NULL){
    sprintf(buffer,"Nebyla zadana cilova adresa");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  np->host=strdup(np->host); 
  if(np->host==NULL){
    LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
    sprintf(buffer,"Nepodarilo se alokovat pamet");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }

  value=parCaseValue(opts,PAGE_TTL);
  if(value!=NULL){
    np->ttl=atoi(value);
    if(np->ttl<=PING_TTL_MIN_VAL||np->ttl>PING_TTL_MAX_VAL){
      sprintf(buffer,"Hodnota TTL musi byt v rozmezi %d-%d",
		PING_TTL_MIN_VAL+1,PING_TTL_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
  }
  value=parCaseValue(opts,PAGE_REPEAT);
  if(value!=NULL){
    np->replies=atoi(value);
    if(np->replies<=PING_REPLIES_MIN_VAL){
      sprintf(buffer,"Hodnota poctu opakovani musi byt vetsi nez %d",
		PING_REPLIES_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef PING_REPLIES_MAX_VAL
    if(np->replies>PING_REPLIES_MAX_VAL){
      sprintf(buffer,"Hodnota poctu opakovani nesmi byt vetsi nez %d",
		PING_REPLIES_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_WAIT);
  if(value!=NULL){
    np->wait_ms=atoi(value);
    if(np->wait_ms<PING_WAIT_MIN_VAL){
      sprintf(buffer,"Prodleva nesmi byt mensi nez %d",PING_WAIT_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef PING_WAIT_MAX_VAL
    if(np->wait_ms>PING_WAIT_MAX_VAL){
      sprintf(buffer,"Prodleva nesmi byt vetsi nez %d",PING_WAIT_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SIZE);
  if(value!=NULL){
    np->pack_size=atoi(value);
    if(np->pack_size<MIN_PING_SIZE){
      sprintf(buffer,"Velikost nesmi byt mensi nez %d",MIN_PING_SIZE);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef PING_SIZE_MAX_VAL
    if(np->pack_size>PING_SIZE_MAX_VAL){
      sprintf(buffer,"Velikost nesmi byt vetsi nez %d",PING_SIZE_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_PATTERN);
  if(value!=NULL){
    np->pattern=strdup(value); 
    if(np->pattern==NULL){
      LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
      sprintf(buffer,"Nepodarilo se alokovat pamet");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
  }
  if (ping_check_params(np) < 0) {
    ping_free(np, NULL);
    np = NULL;
    return -2;
  }
  pstat=ping_exec(np,socketFile);
  ping_rtt_stats(np,pstat,socketFile);
  fflush(socketFile);
  ping_free(np,pstat);
  return 0;
}



/*
 * Volani funkce pro TCP testovani a overeni hodnot z formulare
 */
int tcpFunction(FILE *socketFile,options *opts){
  char buffer[OPT_MAX_LEN];
  char *value;
  parametry * np = (parametry *) malloc (sizeof(parametry));
  *np = def_tcpudp;
  np->IP=parCaseValue(opts,PAGE_ADDRESS);
  if(np->IP==NULL){
    sprintf(buffer,"Nebyla zadana cilova adresa");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  np->IP=all_to_IP(np->IP);
  if(np->IP==NULL){
    sprintf(buffer,"Adresa nebyla nalezena");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  value=parCaseValue(opts,PAGE_PORT);
  if(value!=NULL){
    np->port=atoi(value);
    if(np->port<=TCP_PORT_MIN_VAL||np->port>TCP_PORT_MAX_VAL){
      sprintf(buffer,"Hodnota portu musi byt v rozmezi %d-%d",
             TCP_PORT_MIN_VAL+1,TCP_PORT_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
  }
  value=parCaseValue(opts,PAGE_REPEAT);
  if(value!=NULL){
    np->count=atoi(value);
    if(np->count<=TCP_REPLIES_MIN_VAL){
      sprintf(buffer,"Hodnota poctu opakovani musi byt vetsi nez %d",
             TCP_REPLIES_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef TCP_REPLIES_MAX_VAL
    if(np->count>TCP_REPLIES_MAX_VAL){
      sprintf(buffer,"Hodnota poctu opakovani nesmi byt vetsi nez %d",
		TCP_REPLIES_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SIZE);
  if(value!=NULL){
    np->length=atoi(value);
    if(np->length<TCP_SIZE_MIN_VAL){
      sprintf(buffer,"Velikost nesmi byt mensi nez %d",TCP_SIZE_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef TCP_SIZE_MAX_VAL
    if(np->length>TCP_SIZE_MAX_VAL){
      sprintf(buffer,"Velikost nesmi byt vetsi nez %d",TCP_SIZE_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SOCKET_BUFFER);
  if(value!=NULL){
    np->buffSocket=atoi(value);
    if(np->buffSocket<=TCP_SOCKET_MIN_VAL){
      sprintf(buffer,"Velikost buferu musi byt vetsi nez %d",
		TCP_SOCKET_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
/*
    if(np->buffSocket<np->length){
      sprintf(buffer,"Velikost buferu nesmi byt mensi nez velikost paketu");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
*/
#ifdef TCP_SOCKET_MAX_VAL
    if(np->buffSocket>TCP_SOCKET_MAX_VAL){
      sprintf(buffer,"Velikost buferu nesmi byt vetsi nez %d",
		TCP_SOCKET_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_PATTERN);
  if(value!=NULL){
    np->vstup_data=value;
  }
  protocolTCP(np, socketFile); 
  fflush(socketFile);
  free(np->IP);
  free(np);
  return 0;
}


/*
 * Volani funkce pro UDP testovani a overeni hodnot z formulare
 */
int udpFunction(FILE *socketFile,options *opts){
  char buffer[OPT_MAX_LEN];
  char *value;
  parametry * np = (parametry *) malloc (sizeof(parametry));
  *np = def_tcpudp;
  np->IP=parCaseValue(opts,PAGE_ADDRESS);
  if(np->IP==NULL){
    sprintf(buffer,"Nebyla zadana cilova adresa");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  np->IP=all_to_IP(np->IP);
  if(np->IP==NULL){
    sprintf(buffer,"Adresa nebyla nalezena");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  value=parCaseValue(opts,PAGE_PORT);
  if(value!=NULL){
    np->port=atoi(value);
    if(np->port<=UDP_PORT_MIN_VAL||np->port>UDP_PORT_MAX_VAL){
      sprintf(buffer,"Hodnota portu musi byt v rozmezi %d-%d",
             UDP_PORT_MIN_VAL+1,UDP_PORT_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
  }
  value=parCaseValue(opts,PAGE_REPEAT);
  if(value!=NULL){
    np->count=atoi(value);
    if(np->count<=UDP_REPLIES_MIN_VAL){
      sprintf(buffer,"Hodnota poctu opakovani musi byt vetsi nez %d",
             UDP_REPLIES_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef UDP_REPLIES_MAX_VAL
    if(np->count>UDP_REPLIES_MAX_VAL){
      sprintf(buffer,"Hodnota poctu opakovani nesmi byt vetsi nez %d",
		UDP_REPLIES_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SIZE);
  if(value!=NULL){
    np->length=atoi(value);
    if(np->length<UDP_SIZE_MIN_VAL){
      sprintf(buffer,"Velikost nesmi byt mensi nez %d",UDP_SIZE_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef UDP_SIZE_MAX_VAL
    if(np->length>UDP_SIZE_MAX_VAL){
      sprintf(buffer,"Velikost nesmi byt vetsi nez %d",UDP_SIZE_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SOCKET_BUFFER);
  if(value!=NULL){
    np->buffSocket=atoi(value);
    if(np->buffSocket<=UDP_SOCKET_MIN_VAL){
      sprintf(buffer,"Velikost buferu musi byt vetsi nez %d",
		UDP_SOCKET_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
    if(np->buffSocket<np->length){
      sprintf(buffer,"Velikost buferu nesmi byt mensi nez velikost paketu");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef UDP_SOCKET_MAX_VAL
    if(np->buffSocket>UDP_SOCKET_MAX_VAL){
      sprintf(buffer,"Velikost buferu nesmi byt vetsi nez %d",
		UDP_SOCKET_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_PATTERN);
  if(value!=NULL){
    np->vstup_data=value;
  }
  protocolUDP2(np, socketFile); 
  fflush(socketFile);
  free(np->IP);
  free(np);
  return 0;
}


/*
 * Volani funkce pro ICMP testovani a overeni hodnot z formulare
 */
int icmpFunction(FILE *socketFile,options *opts){
  char buffer[OPT_MAX_LEN];
  char *value;
  parametry * np = (parametry *) malloc (sizeof(parametry));
  *np = def_tcpudp;
  np->IP=parCaseValue(opts,PAGE_ADDRESS);
  if(np->IP==NULL){
    sprintf(buffer,"Nebyla zadana cilova adresa");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  np->IP=all_to_IP(np->IP);
  if(np->IP==NULL){
    sprintf(buffer,"Adresa nebyla nalezena");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  value=parCaseValue(opts,PAGE_REPEAT);
  if(value!=NULL){
    np->count=atoi(value);
    if(np->count<=ICMP_REPLIES_MIN_VAL){
      sprintf(buffer,"Hodnota poctu opakovani musi byt vetsi nez %d",
             ICMP_REPLIES_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef ICMP_REPLIES_MAX_VAL
    if(np->count>ICMP_REPLIES_MAX_VAL){
      sprintf(buffer,"Hodnota poctu opakovani nesmi byt vetsi nez %d",
		ICMP_REPLIES_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_WAIT);
  if(value!=NULL){
    np->freq=atoi(value);
    if(np->freq<ICMP_WAIT_MIN_VAL){
      sprintf(buffer,"Prodleva nesmi byt mensi nez %d",ICMP_WAIT_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef ICMP_WAIT_MAX_VAL
    if(np->freq>ICMP_WAIT_MAX_VAL){
      sprintf(buffer,"Prodleva nesmi byt vetsi nez %d",ICMP_WAIT_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SIZE);
  if(value!=NULL){
    np->length=atoi(value);
    if(np->length<ICMP_SIZE_MIN_VAL){
      sprintf(buffer,"Velikost nesmi byt mensi nez %d",ICMP_SIZE_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef ICMP_SIZE_MAX_VAL
    if(np->length>ICMP_SIZE_MAX_VAL){
      sprintf(buffer,"Velikost nesmi byt vetsi nez %d",ICMP_SIZE_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SOCKET_BUFFER);
  if(value!=NULL){
    np->buffSocket=atoi(value);
    if(np->buffSocket<=ICMP_SOCKET_MIN_VAL){
      sprintf(buffer,"Velikost buferu musi byt vetsi nez %d",
		ICMP_SOCKET_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
/*
    if(np->buffSocket<np->length){
      sprintf(buffer,"Velikost buferu nesmi byt mensi nez velikost paketu");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
*/
#ifdef ICMP_SOCKET_MAX_VAL
    if(np->buffSocket>ICMP_SOCKET_MAX_VAL){
      sprintf(buffer,"Velikost buferu nesmi byt vetsi nez %d",
		ICMP_SOCKET_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_PATTERN);
  if(value!=NULL){
    np->vstup_data=value;
  }
  protocolICMP(np, socketFile); 
  free(np->IP);
  fflush(socketFile);
  free(np);
  return 0;
}


/*
 * Volani funkce pro traceroute a overeni hodnot z formulare
 */
int traceFunction(FILE *socketFile,options *opts){
  char buffer[OPT_MAX_LEN];
  char *value;
  TRACER_STATS *pstat;
  TRACER * np = (TRACER *) malloc (sizeof(TRACER));
  *np = def_tr;
  np->host=parCaseValue(opts,PAGE_ADDRESS);
  if(np->host==NULL){
    sprintf(buffer,"Nebyla zadana cilova adresa");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  np->host=strdup(np->host); 
  if(np->host==NULL){
    LOG_ERR("HTTP","Nepodarilo se alokovat pamet");
    sprintf(buffer,"Nepodarilo se alokovat pamet");
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }

  value=parCaseValue(opts,PAGE_TTL);
  if(value!=NULL){
    np->max_ttl=atoi(value);
    if(np->max_ttl<=TRACE_TTL_MIN_VAL||np->max_ttl>TRACE_TTL_MAX_VAL){
      sprintf(buffer,"Hodnota TTL musi byt v rozmezi %d-%d",
		TRACE_TTL_MIN_VAL+1,TRACE_TTL_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
  }
  value=parCaseValue(opts,PAGE_MIN_TTL);
  if(value!=NULL){
    np->min_ttl=atoi(value);
    if(np->min_ttl<=TRACE_TTLMIN_MIN_VAL||np->min_ttl>TRACE_TTLMIN_MAX_VAL){
      sprintf(buffer,"Minimalni hodnota TTL musi byt v rozmezi %d-%d",
             TRACE_TTLMIN_MIN_VAL+1,TRACE_TTLMIN_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
    else if(np->min_ttl>np->max_ttl){
      sprintf(buffer,"Minimalni hodnota nesmi byt vetsi nez maximalni");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
  }
  value=parCaseValue(opts,PAGE_REPEAT);
  if(value!=NULL){
    np->queries=atoi(value);
    if(np->queries<=TRACE_REPLIES_MIN_VAL){
      sprintf(buffer,"Hodnota poctu opakovani musi byt vetsi nez %d",
		TRACE_REPLIES_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef TRACE_REPLIES_MAX_VAL
    if(np->queries>TRACE_REPLIES_MAX_VAL){
      sprintf(buffer,"Hodnota poctu opakovani nesmi byt vetsi nez %d",
		TRACE_REPLIES_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_WAIT);
  if(value!=NULL){
    np->wait_ms=atoi(value);
    if(np->wait_ms<TRACE_WAIT_MIN_VAL){
      sprintf(buffer,"Prodleva nesmi byt mensi nez %d",TRACE_WAIT_MIN_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef TRACE_WAIT_MAX_VAL
    if(np->wait_ms>TRACE_WAIT_MAX_VAL){
      sprintf(buffer,"Prodleva nesmi byt vetsi nez %d",TRACE_WAIT_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  value=parCaseValue(opts,PAGE_SIZE);
  if(value!=NULL){
    np->pack_size=atoi(value);
    if(np->pack_size<MIN_PING_SIZE){
      sprintf(buffer,"Velikost nesmi byt mensi nez %d",MIN_PING_SIZE);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#ifdef TRACE_SIZE_MAX_VAL
    if(np->pack_size>TRACE_SIZE_MAX_VAL){
      sprintf(buffer,"Velikost nesmi byt vetsi nez %d",TRACE_SIZE_MAX_VAL);
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
#endif
  }
  if (tracert_check_params(np) < 0) {
    tracert_free(np, NULL);
    np = NULL;
    return -2;
  }
  pstat=tracert_exec(np,socketFile);
  fflush(socketFile);
  tracert_free(np,pstat);
  return 0;
}

/*
 * Vybere spravnou funkci podle hodnot z formulare
 */
int chooseFunction(FILE *socketFile,options *opts){
  char buffer[OPT_MAX_LEN];
  char *val;
  val=parCaseValue(opts,FUNCTION);
  if(val==NULL){
    sprintf(buffer,"Ve formulari chybi urceni '%s'",FUNCTION);
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
  if(strcmp(val, PING_FUNCTION)==0){
    if(pingFunction(socketFile,opts)==-2){
      sprintf(buffer,"Spatne parametry");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
    return 0;
  }
  else if(strcmp(val, TRACE_FUNCTION)==0){
    if(traceFunction(socketFile,opts)==-2){
      sprintf(buffer,"Spatne parametry");
      fwrite(buffer,strlen(buffer),1,socketFile);
      return -1;
    }
    return 0;
  }
  else if(strcmp(val, TCP_FUNCTION)==0){
    tcpFunction(socketFile,opts);
    return 0;
  }
  else if(strcmp(val, UDP_FUNCTION)==0){
    udpFunction(socketFile,opts);
    return 0;
  }
  else if(strcmp(val, ICMP_FUNCTION)==0){
    icmpFunction(socketFile,opts);
    return 0;
  }
  else{
    sprintf(buffer,"Pozadovana neznama funkce '%s'",val);
    fwrite(buffer,strlen(buffer),1,socketFile);
    return -1;
  }
}

/*
 * Zapise hlavicku generovanych dat, spusti prislusnou funkci a potom zapise
 * konec HTML stranky
 */
void analyze(FILE *socketFile,request *req){
  char buffer[OPT_MAX_LEN];
  int i;
  options *opts;
  req->mimeType="text/html";
  writeHeader(socketFile,req, HTTP_OK,0,"Expires: 0\r\nPragma: no-cache\r\n");
  opts=parseParams(req->params);
  sprintf(buffer,"<HTML><HEAD><TITLE>Vystup programu</TITLE>\n"
    "<meta http-equiv=\"Content-Type\" content=\"text/html; "
    "charset=ISO 8859-2\"></HEAD>\n"
    "<BODY>");
#ifdef PRINT_OPTS_VALS
  for(i=0;i<opts->count;i++){
     printf("%s - %s\n",
	opts->opt[i].name,opts->opt[i].value);
  }
#endif
  if(req->type!=HEAD){
    fwrite(buffer,strlen(buffer),1,socketFile);
    chooseFunction(socketFile,opts);
    sprintf(buffer,"</BODY></HTML>");
    fwrite(buffer,strlen(buffer),1,socketFile);
  }
  for(i=0;i<opts->count;i++){
    if(opts->opt[i].value!=NULL){
      free(opts->opt[i].value);
    }
    if(opts->opt[i].name!=NULL){
      free(opts->opt[i].name);
    }
  }
  if(opts->opt!=NULL){ 
    free(opts->opt);
  }
  free(opts);
}
